"""
Module with useful time/date functions
"""

import datetime
import dateutil.parser


def get_dates_in_range(start_date, end_date):
    """ Get all dates within input start and end date in ISO 8601 format

    :param start_date: start date in ISO 8601 format
    :type start_date: str
    :param end_date: end date in ISO 8601 format
    :type end_date: str
    :return: list of dates between start_date and end_date in ISO 8601 format
    :rtype: list of str
    """
    start_dt = iso_to_datetime(start_date)
    end_dt = iso_to_datetime(end_date)
    num_days = int((end_dt - start_dt).days)
    return [datetime_to_iso(start_dt + datetime.timedelta(i)) for i in range(num_days + 1)]


def next_date(date):
    """ Get date of day after input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-13'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date after input date in ISO 8601 format
    :rtype: str
    """
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm + datetime.timedelta(1))


def prev_date(date):
    """ Get date of day previous to input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-11'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date previous to input date in ISO 8601 format
    :rtype: str
    """
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm - datetime.timedelta(1))


def iso_to_datetime(date):
    """ Convert ISO 8601 time format to datetime format

    This function converts a date in ISO format, e.g. ``2017-09-14`` to a ``datetime`` instance, e.g.
    ``datetime.datetime(2017,9,14,0,0)``

    :param date: date in ISO 8601 format
    :type date: str
    :return: datetime instance
    :rtype: datetime
    """
    chunks = list(map(int, date.split('T')[0].split('-')))
    return datetime.datetime(chunks[0], chunks[1], chunks[2])


def datetime_to_iso(date, only_date=True):
    """ Convert datetime format to ISO 8601 time format

    This function converts a date in datetime instance, e.g. ``datetime.datetime(2017,9,14,0,0)`` to ISO format,
    e.g. ``2017-09-14``

    :param date: datetime instance to convert
    :type date: datetime
    :param only_date: whether to return date only or also time information. Default is ``True``
    :type only_date: bool
    :return: date in ISO 8601 format
    :rtype: str
    """
    if only_date:
        return date.isoformat().split('T')[0]
    return date.isoformat()


def get_current_date():
    """ Get current date in ISO 8601 format

    :return: current date in ISO 8601 format
    :rtype: str
    """
    date = datetime.datetime.now()
    return datetime_to_iso(date)


def is_valid_time(time):
    """ Check if input string represents a valid time/date stamp

    :param time: a string containing a time/date stamp
    :type time: str
    :return: ``True`` is string is a valid time/date stamp, ``False`` otherwise
    :rtype: bool
    """
    try:
        dateutil.parser.parse(time)
        return True
    except BaseException:
        return False


def parse_time(time_str):
    """ Parse input time/date string as ISO 8601 string

    :param time_str: time/date string to parse
    :type time_str: str
    :return: parsed string in ISO 8601 format
    :rtype: str
    """
    if len(time_str) < 8:
        raise ValueError('Invalid time string {}.\n'
                         'Please specify time in formats YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS'.format(time_str))
    time = dateutil.parser.parse(time_str)
    if len(time_str) <= 10:
        return time.date().isoformat()
    return time.isoformat()



"""
This module lists all externally useful classes and functions
"""

from .data_request import WmsRequest, WcsRequest, AwsTileRequest, AwsProductRequest, GeopediaWmsRequest, \
    get_safe_format, download_safe_format

from .aws import AwsProduct, AwsTile
from .aws_safe import SafeProduct, SafeTile

from .areas import BBoxSplitter, OsmSplitter, TileSplitter

from .ogc import WebFeatureService

from .common import BBox
from .constants import DataSource, CustomUrlParam, CRS, MimeType, OgcConstants, AwsConstants, ServiceType
from .config import SHConfig

from .download import DownloadRequest, download_data, get_json, get_xml, DownloadFailedException, \
    AwsDownloadFailedException

from .opensearch import get_tile_info_id, get_tile_info, get_area_dates, get_area_info

from .io_utils import read_data, write_data
from .os_utils import get_content_list, get_folder_list, get_file_list, make_folder, create_parent_folder, rename, size
from .geo_utils import bbox_to_resolution, get_image_dimension, to_utm_bbox, get_utm_bbox, wgs84_to_utm, to_wgs84, \
    utm_to_pixel, pixel_to_utm, wgs84_to_pixel, get_utm_crs, transform_point, transform_bbox
from .time_utils import next_date, prev_date, get_current_date


__version__ = "2.1.0"



"""
Module for working with Sentinel Hub OGC services
"""

import logging
import datetime

from base64 import b64encode
from urllib.parse import urlencode
from shapely.geometry import shape as geo_shape

from .time_utils import get_current_date, parse_time
from .download import DownloadRequest, get_json
from .constants import ServiceType, DataSource, MimeType, CRS, OgcConstants, CustomUrlParam
from .config import SHConfig
from .geo_utils import get_image_dimension

LOGGER = logging.getLogger(__name__)


class OgcService:
    """ The base class for Sentinel Hub OGC services

    :param base_url: base url of Sentinel Hub's OGC services. If ``None``, the url specified in the configuration
                    file is taken.
    :type base_url: str or None
    :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If ``None``, the instance id
                        specified in the configuration file is taken.
    :type instance_id: str or None
    """
    def __init__(self, base_url=None, instance_id=None):
        self.base_url = SHConfig().ogc_base_url if not base_url else base_url
        self.instance_id = SHConfig().instance_id if not instance_id else instance_id

        if not self.instance_id:
            raise ValueError('Instance ID is not set. '
                             'Set it either in request initialisation or in configuration file.')

    @staticmethod
    def _parse_time_interval(time):
        """ Parses times into common form

        Parses specified time into common form - tuple of start and end dates, i.e.:

        ``(2017-01-15:T00:00:00, 2017-01-16:T23:59:59)``

        The parameter can have the following values/format, which will be parsed as:

        * ``None`` -> `[default_start_date from config.json, current date]`
        * `YYYY-MM-DD` -> `[YYYY-MM-DD:T00:00:00, YYYY-MM-DD:T23:59:59]`
        * `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`
        * list or tuple of two dates (`YYYY-MM-DD`) -> `[YYYY-MM-DDT00:00:00, YYYY-MM-DDT23:59:59]`, where the first
          (second) element is start (end) date
        * list or tuple of two dates (`YYYY-MM-DDThh:mm:ss`) -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`,
          where the first (second) element is start (end) date

        :param time: time window of acceptable acquisitions. See above for all acceptable argument formats.
        :type time: ``None``, str of form `YYYY-MM-DD` or `'YYYY-MM-DDThh:mm:ss'`, list or tuple of two such strings
        :return: interval of start and end date of the form YYYY-MM-DDThh:mm:ss
        :rtype: tuple of start and end date
        """
        if time is None or time is OgcConstants.LATEST:
            date_interval = (SHConfig().default_start_date, get_current_date())
        else:
            if isinstance(time, str):
                date_interval = (parse_time(time), parse_time(time))
            elif isinstance(time, list) or isinstance(time, tuple) and len(time) == 2:
                date_interval = (parse_time(time[0]), parse_time(time[1]))
            else:
                raise TabError('time must be a string or tuple of 2 strings or list of 2 strings')
            if date_interval[0] > date_interval[1]:
                raise ValueError('First time must be smaller or equal to second time')

        if len(date_interval[0].split('T')) == 1:
            date_interval = (date_interval[0] + 'T00:00:00', date_interval[1])
        if len(date_interval[1].split('T')) == 1:
            date_interval = (date_interval[0], date_interval[1] + 'T23:59:59')

        return date_interval

    @staticmethod
    def _filter_dates(dates, time_difference):
        """
        Filters out dates within time_difference, preserving only the oldest date.

        :param dates: a list of datetime objects
        :param time_difference: a ``datetime.timedelta`` representing the time difference threshold
        :return: an ordered list of datetimes `d1<=d2<=...<=dn` such that `d[i+1]-di > time_difference`
        :rtype: list[datetime.datetime]
        """

        LOGGER.debug("dates=%s", dates)

        if len(dates) <= 1:
            return dates

        sorted_dates = sorted(dates)

        separate_dates = [sorted_dates[0]]
        for curr_date in sorted_dates[1:]:
            if curr_date - separate_dates[-1] > time_difference:
                separate_dates.append(curr_date)
        return separate_dates

    @staticmethod
    def _sentinel1_product_check(product_id, data_source):
        """Checks if Sentinel-1 product ID matches Sentinel-1 DataSource configuration

        :param product_id: Sentinel-1 product ID
        :type product_id: str
        :param data_source: One of the supported Sentinel-1 data sources
        :type data_source: constants.DataSource
        :return: True if data_source contains product_id and False otherwise
        :rtype: bool
        """
        props = product_id.split('_')
        acquisition, resolution, polarisation = props[1], props[2][3], props[3][2:4]
        if acquisition in ['IW', 'EW'] and resolution in ['M', 'H'] and polarisation in ['DV', 'DH', 'SV', 'SH']:
            return acquisition == data_source.value[2].name and polarisation == data_source.value[3].name and \
                   resolution == data_source.value[4].name[0]
        raise ValueError('Unknown Sentinel-1 tile type: {}'.format(product_id))


class OgcImageService(OgcService):
    """Sentinel Hub OGC services class for providing image data

    Intermediate layer between QGC-type requests (WmsRequest and WcsRequest) and the Sentinel Hub OGC (WMS and WCS)
    services.

    :param base_url: base url of Sentinel Hub's OGC services. If ``None``, the url specified in the configuration
                    file is taken.
    :type base_url: str or None
    :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If ``None``, the instance id
                        specified in the configuration file is taken.
    :type instance_id: str or None
    """
    def __init__(self, **kwargs):
        super(OgcImageService, self).__init__(**kwargs)

        self.wfs_iterator = None

    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: QGC-type request with specified bounding box, time interval, and cloud coverage for specific
                        product.
        :type request: OgcRequest or GeopediaRequest
        :return: list of DownloadRequests
        """
        size_x, size_y = self.get_image_dimensions(request)
        return [DownloadRequest(url=self.get_url(request, date, size_x, size_y),
                                filename=self.get_filename(request, date, size_x, size_y),
                                data_type=request.image_format, headers=OgcConstants.HEADERS)
                for date in self.get_dates(request)]

    def get_url(self, request, date, size_x, size_y):
        """ Returns url to Sentinel Hub's OGC service for the product specified by the OgcRequest and date.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return: url to Sentinel Hub's OGC service for this product.
        :rtype: str
        """
        url = self.base_url + request.service_type.value
        # These 2 lines are temporal and will be removed after the use of uswest url wont be required anymore:
        if hasattr(request, 'data_source') and DataSource.is_uswest_source(request.data_source):
            url = 'https://services-uswest2.sentinel-hub.com/ogc/{}'.format(request.service_type.value)

        params = {'SERVICE': request.service_type.value,
                  'BBOX': request.bbox.__str__(reverse=True) if request.bbox.get_crs() is CRS.WGS84 else str(
                      request.bbox),
                  'FORMAT': MimeType.get_string(request.image_format),
                  'CRS': CRS.ogc_string(request.bbox.get_crs())}

        if request.service_type is ServiceType.WMS:
            params = {**params,
                      **{
                          'WIDTH': size_x,
                          'HEIGHT': size_y,
                          'LAYERS': request.layer,
                          'REQUEST': 'GetMap'
                      }}
        elif request.service_type is ServiceType.WCS:
            params = {**params,
                      **{
                          'RESX': size_x,
                          'RESY': size_y,
                          'COVERAGE': request.layer,
                          'REQUEST': 'GetCoverage'
                      }}

        if date is not None:
            start_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date - request.time_difference
            end_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date + request.time_difference
            params['TIME'] = '{}/{}'.format(start_date.isoformat(), end_date.isoformat())

        if hasattr(request, 'maxcc'):
            params['MAXCC'] = 100.0 * request.maxcc

        if hasattr(request, 'custom_url_params') and request.custom_url_params is not None:
            params = {**params,
                      **{k.value: str(v) for k, v in request.custom_url_params.items()}}

            if CustomUrlParam.EVALSCRIPT.value in params:
                evalscript = params[CustomUrlParam.EVALSCRIPT.value]
                params[CustomUrlParam.EVALSCRIPT.value] = b64encode(evalscript.encode()).decode()

        authority = self.instance_id if hasattr(self, 'instance_id') else request.theme
        return '{}/{}?{}'.format(url, authority, urlencode(params))

    @staticmethod
    def get_filename(request, date, size_x, size_y):
        """ Get filename location

        Returns the filename's location on disk where data is or is going to be stored.
        The files are stored in the folder specified by the user when initialising OGC-type
        of request. The name of the file has the following structure:

        {service_type}_{layer}_{crs}_{bbox}_{time}_{size_x}X{size_y}_{custom_url_param}_
        {custom_url_param_val}.{image_format}

        In case of `TIFF_d32f` a `'_tiff_depth32f'` is added at the end of the filename (before format suffix)
        to differentiate it from 16-bit float tiff.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return: filename for this request and date
        :rtype: str
        """
        filename = '_'.join([str(request.service_type.value),
                             request.layer,
                             str(request.bbox.get_crs()).replace(':', ''),
                             str(request.bbox).replace(',', '_'),
                             '' if date is None else date.strftime("%Y-%m-%dT%H-%M-%S"),
                             '{}X{}'.format(size_x, size_y)])

        LOGGER.debug("filename=%s", filename)

        if hasattr(request, 'custom_url_params') and request.custom_url_params is not None:
            for param, value in sorted(request.custom_url_params.items(),
                                       key=lambda parameter_item: parameter_item[0].value):
                filename = '_'.join([filename, param.value, str(value)])

        for char in [' ', '/', '\\', '|', ';', ':', '\n', '\t']:
            filename = filename.replace(char, '')

        suffix = str(request.image_format.value)
        if request.image_format is MimeType.TIFF_d32f:
            suffix = str(MimeType.TIFF.value)
            filename = '_'.join([filename, str(request.image_format.value).replace(';', '_')])

        filename = '.'.join([filename[:254 - len(suffix)], suffix])
        return filename   # Even in UNIX systems filename must have at most 255 bytes

    def get_dates(self, request):
        """ Get available Sentinel-2 acquisitions at least time_difference apart

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        When a time_difference threshold is set to a positive value, the function filters out all datetimes which
        are within the time difference. The oldest datetime is preserved, all others all deleted.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: List of dates of existing acquisitions for the given request
        :rtype: list(datetime.datetime) or [None]
        """
        if DataSource.is_timeless(request.data_source):
            return [None]

        date_interval = OgcService._parse_time_interval(request.time)

        LOGGER.debug('date_interval=%s', date_interval)

        self.wfs_iterator = WebFeatureService(request.bbox, date_interval, data_source=request.data_source,
                                              maxcc=request.maxcc, base_url=self.base_url, instance_id=self.instance_id)

        dates = sorted(set(self.wfs_iterator.get_dates()))

        if request.time is OgcConstants.LATEST:
            dates = dates[-1:]
        return OgcService._filter_dates(dates, request.time_difference)

    @staticmethod
    def get_image_dimensions(request):
        """
        Verifies or calculates image dimensions.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: horizontal and vertical dimensions of requested image
        :rtype: (int or str, int or str)
        """
        if request.service_type is ServiceType.WCS or (isinstance(request.size_x, int) and
                                                       isinstance(request.size_y, int)):
            return request.size_x, request.size_y
        if not isinstance(request.size_x, int) and not isinstance(request.size_y, int):
            raise ValueError("At least one of parameters 'width' and 'height' must have an integer value")
        missing_dimension = get_image_dimension(request.bbox, width=request.size_x, height=request.size_y)
        if request.size_x is None:
            return missing_dimension, request.size_y
        if request.size_y is None:
            return request.size_x, missing_dimension
        raise ValueError("Parameters 'width' and 'height' must be integers or None")

    def get_wfs_iterator(self):
        """Returns iterator over info about all satellite tiles used for the request

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
                 DataSource.DEM it returns None.
        :rtype: Iterator[dict] or None
        """
        return self.wfs_iterator


class WebFeatureService(OgcService):
    """Class for interaction with Sentinel Hub WFS service

    The class is an iterator over info data of all available satellite tiles for requested parameters. It collects data
    from Sentinel Hub service only during the first iteration. During next iterations it returns already obtained data.
    The data is in the order returned by Sentinel Hub WFS service.

    :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference system.
    :type bbox: common.BBox
    :param time_interval: interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
    :type time_interval: (str, str)
    :param data_source: Source of requested satellite data. Default is Sentinel-2 L1C data.
    :type data_source: constants.DataSource
    :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
    :type maxcc: float
    :param base_url: base url of Sentinel Hub's OGC services. If ``None``, the url specified in the configuration
                    file is taken.
    :type base_url: str or None
    :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If ``None``, the instance id
                        specified in the configuration file is taken.
    :type instance_id: str or None
    """
    def __init__(self, bbox, time_interval, *, data_source=DataSource.SENTINEL2_L1C, maxcc=1.0, **kwargs):
        super(WebFeatureService, self).__init__(**kwargs)

        self.bbox = bbox
        self.time_interval = self._parse_time_interval(time_interval)
        self.data_source = data_source
        self.maxcc = maxcc

        self.tile_list = []
        self.index = 0
        self.feature_offset = 0

    def __iter__(self):
        """Iteration method

        :return: iterator of dictionaries containing info about product tiles
        :rtype: Iterator[dict]
        """
        self.index = 0
        return self

    def __next__(self):
        """Next method

        :return: dictionary containing info about product tiles
        :rtype: dict
        """
        while self.index >= len(self.tile_list) and self.feature_offset is not None:
            self._fetch_features()

        if self.index < len(self.tile_list):
            self.index += 1
            return self.tile_list[self.index - 1]

        raise StopIteration

    def _fetch_features(self):
        """Collects data from WFS service

        :return: dictionary containing info about product tiles
        :rtype: dict
        """
        if self.feature_offset is None:
            return

        main_url = '{}{}/{}?'.format(self.base_url, ServiceType.WFS.value, self.instance_id)

        params = {'SERVICE': ServiceType.WFS.value,
                  'REQUEST': 'GetFeature',
                  'TYPENAMES': DataSource.get_wfs_typename(self.data_source),
                  'BBOX': self.bbox.__str__(reverse=True) if self.bbox.get_crs() is CRS.WGS84 else str(self.bbox),
                  'OUTPUTFORMAT': MimeType.get_string(MimeType.JSON),
                  'SRSNAME': CRS.ogc_string(self.bbox.get_crs()),
                  'TIME': '{}/{}'.format(self.time_interval[0], self.time_interval[1]),
                  'MAXCC': 100.0 * self.maxcc,
                  'MAXFEATURES': SHConfig().max_wfs_records_per_query,
                  'FEATURE_OFFSET': self.feature_offset}

        url = main_url + urlencode(params)
        LOGGER.debug("URL=%s", url)
        response = get_json(url)

        is_sentinel1 = DataSource.is_sentinel1(self.data_source)
        for tile_info in response["features"]:
            if not is_sentinel1 or self._sentinel1_product_check(tile_info['properties']['id'], self.data_source):
                self.tile_list.append(tile_info)

        if len(response["features"]) < SHConfig().max_wfs_records_per_query:
            self.feature_offset = None
        else:
            self.feature_offset += SHConfig().max_wfs_records_per_query

    def get_dates(self):
        """ Returns a list of acquisition times from tile info data

        :return: List of acquisition times in the order returned by WFS service.
        :rtype: list(datetime.datetime)
        """
        return [datetime.datetime.strptime('{}T{}'.format(tile_info['properties']['date'],
                                                          tile_info['properties']['time'].split('.')[0]),
                                           '%Y-%m-%dT%H:%M:%S') for tile_info in self]

    def get_geometries(self):
        """ Returns a list of geometries from tile info data

        :return: List of multipolygon geometries in the order returned by WFS service.
        :rtype: list(shapely.geometry.MultiPolygon)
        """
        return [geo_shape(tile_info['geometry']) for tile_info in self]

    def get_tiles(self):
        """ Returns list of tiles with tile name, date and AWS index

        :return: List of tiles in form of (tile_name, date, aws_index)
        :rtype: list((str, str, int))
        """
        return [self._parse_tile_url(tile_info['properties']['path']) for tile_info in self]

    @staticmethod
    def _parse_tile_url(tile_url):
        """ Extracts tile name, data and AWS index from tile URL

        :param tile_url: Location of tile at AWS
        :type: tile_url: str
        :return: Tuple in a form (tile_name, date, aws_index)
        :rtype: (str, str, int)
        """
        props = tile_url.rsplit('/', 7)
        return ''.join(props[1:4]), '-'.join(props[4:7]), int(props[7])



"""
Module that collects configuration data from config.json
"""

import os.path
import json
from collections import OrderedDict

# pylint: disable=E0203


class SHConfig:
    """ This is a singleton implementation of the sentinelhub configuration class.

    The class reads during its first initialisation the configurable settings from
    ``./config.json`` file:

        - `instance_id`: Users' instance id. User can set it to his/hers instance id in ``config.json`` instead
          of specifying it explicitly every time he/she creates new ogc request.
        - `aws_access_key_id`: Access key for AWS Requester Pays buckets.
        - `aws_secret_access_key`: Secret access key for AWS Requester Pays buckets.
        - `ogc_base_url`: Base url for Sentinel Hub's services (should not be changed by the user).
        - `gpd_base_url`: Base url for Geopedia's services (should not be changed by the user).
        - `aws_base_url`: Base url for Sentinel-2 data on AWS (should not be changed by the user).
        - `aws_s3_l1c_bucket`: Name of Sentinel-2 L1C bucket at AWS s3 service.
        - `use_s3_l1c_bucket`: If `true` L1C data will be downloaded from s3 bucket, if `false` it will be downloaded
          from free of charge http source.
        - `aws_s3_l2a_bucket`: Name of Sentinel-2 L2A bucket at AWS s3 service.
        - `opensearch_url`: Base url for Sentinelhub Opensearch service.
        - `max_wfs_records_per_query`: Maximum number of records returned for each WFS query.
        - `max_opensearch_records_per_query`: Maximum number of records returned for each Opensearch query.
        - `default_start_date`: In case time parameter for OGC data requests is not specified this will be used for
          start date of the interval.
        - `max_download_attempts`: Maximum number of download attempts from a single URL until an error will be raised.
        - `download_sleep_time`: Number of seconds between the failed download attempt and the next attempt.
        - `download_timeout_seconds`: Maximum number of seconds before download attempt is canceled.

    Usage in the code:

        * ``SHConfig().ogc_base_url``
        * ``SHConfig().instance_id``

    """
    class _SHConfig:
        """
        Private class.
        """
        def __init__(self):
            self.config_params = ['instance_id', 'aws_access_key_id', 'aws_secret_access_key', 'ogc_base_url',
                                  'gpd_base_url', 'aws_base_url', 'aws_s3_l1c_bucket', 'use_s3_l1c_bucket',
                                  'aws_s3_l2a_bucket', 'opensearch_url', 'max_wfs_records_per_query',
                                  'max_opensearch_records_per_query', 'default_start_date', 'max_download_attempts',
                                  'download_sleep_time', 'download_timeout_seconds']
            self.instance_id = ''
            self.load_configuration()

        def _check_configuration(self, config):
            """
            Checks if configuration file has contains all keys.

            :param config: configuration dictionary read from ``config.json``
            :type config: dict
            """

            for key in self.config_params:
                if key not in config:
                    raise ValueError('Configuration file does not contain %s key.' % key)
            if config['max_wfs_records_per_query'] > 100:
                raise ValueError("Value of config parameter 'max_wfs_records_per_query' must be at most 100")
            if config['max_opensearch_records_per_query'] > 500:
                raise ValueError("Value of config parameter 'max_opensearch_records_per_query' must be at most 500")

        @staticmethod
        def _get_config_file():
            """Checks if configuration file exists and returns its file path

            :return: location of configuration file
            :rtype: str
            """
            package_dir = os.path.dirname(__file__)
            config_file = os.path.join(package_dir, 'config.json')

            if not os.path.isfile(config_file):
                raise IOError('Configuration file does not exist: %s' % os.path.abspath(config_file))

            return config_file

        def load_configuration(self):
            """
            Method reads and loads the configuration file.
            """
            with open(self._get_config_file(), 'r') as cfg_file:
                config = json.load(cfg_file)
                self._check_configuration(config)

                for prop in config:
                    if prop in self.config_params:
                        setattr(self, prop, config[prop])

                if not self.instance_id:
                    self.instance_id = None

        def get_config(self):
            """Returns ordered dictionary with configuration parameters

            :return: Ordered dictionary
            :rtype: collections.OrderedDict
            """
            config = OrderedDict((prop, getattr(self, prop)) for prop in self.config_params)
            if config['instance_id'] is None:
                config['instance_id'] = ''
            return config

        def save_configuration(self):
            """
            Method saves changed parameter values to the configuration file.
            """
            config = self.get_config()

            self._check_configuration(config)

            with open(self._get_config_file(), 'w') as cfg_file:
                json.dump(config, cfg_file, indent=2)

    _instance = None

    def __init__(self):
        if not SHConfig._instance:
            SHConfig._instance = self._SHConfig()

    def __getattr__(self, name):
        return getattr(self._instance, name)

    def __dir__(self):
        return sorted(list(dir(super(SHConfig, self))) + self._instance.config_params)

    def __str__(self):
        return json.dumps(self._instance.get_config(), indent=2)

    def save(self):
        """Method that saves configuration parameter changes from instance of SHConfig class to global config class and
        to ``config.json`` file.

        Example of use case
            ``my_config = SHConfig()`` \n
            ``my_config.instance_id = '<new instance id>'`` \n
            ``my_config.save()``
        """
        for prop in self._instance.config_params:
            setattr(self._instance, prop, getattr(self, prop))
        self._instance.save_configuration()

    def get_params(self):
        """Returns a list of parameter names

        :return: List of parameter names
        :rtype: list(str)
        """
        return self._instance.config_params



"""
Module for downloading data
"""

import logging
import os
import time
import concurrent.futures
import requests
import json
import cv2
import boto3
import warnings
import numpy as np
import tifffile as tiff
from io import BytesIO
from xml.etree import ElementTree

from .os_utils import create_parent_folder, sys_is_windows
from .constants import MimeType, RequestType
from .config import SHConfig
from .io_utils import get_jp2_bit_depth, _fix_jp2_image

LOGGER = logging.getLogger(__name__)


class DownloadFailedException(Exception):
    pass


class AwsDownloadFailedException(DownloadFailedException):
    """
    This exception is raised when download fails because of a missing file in AWS
    """
    pass


class ImageDecodingError(Exception):
    pass


class DownloadRequest:
    """ Class to manage HTTP requests

    Container for all download requests issued by the DataRequests containing
    url to Sentinel Hub's services or other sources to data, file names to
    saved data and other necessary flags needed when the data is downloaded and
    interpreted.

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded. Default is ``None``
    :type url: str
    :param data_folder: folder name where the fetched data will be (or already is) saved. Default is ``None``
    :type data_folder: str
    :param filename: filename of the file where the fetched data will be (or already is) saved. Default is ``None``
    :type filename: str
    :param headers: add HTTP headers to request. Default is ``None``
    :type headers: dict
    :param request_type: type of request, either GET or POST. Default is ``constants.RequestType.GET``
    :type request_type: constants.RequestType
    :param post_values: form encoded data to send in POST request. Default is ``None``
    :type post_values: dict
    :param save_response: flag to turn on/off saving data downloaded by this request to disk. Default is ``True``.
    :type save_response: bool
    :param return_data: flag to return or not data downloaded by this request to disk. Default is ``True``.
    :type return_data: bool
    :param data_type: expected file format of downloaded data. Default is ``constants.MimeType.RAW``
    :type data_type: constants.MimeType

    """
    def __init__(self, *, url=None, data_folder=None, filename=None, headers=None, request_type=RequestType.GET,
                 post_values=None, save_response=True, return_data=True, data_type=MimeType.RAW, **properties):

        self.url = url
        self.data_folder = data_folder
        self.filename = filename
        self.headers = headers
        self.post_values = post_values
        self.save_response = save_response
        self.return_data = return_data

        self.properties = properties

        self.request_type = RequestType(request_type)
        self.data_type = MimeType(data_type)

        self.will_download = True
        self.file_location = None
        self._set_file_location()

    def set_save_response(self, save_response):
        """
        Set save_response attribute

        :param save_response: flag to turn on/off saving data downloaded by this request to disk. Default is ``True``.
        :return: bool
        """
        self.save_response = save_response

    def set_return_data(self, return_data):
        """
        Set return_data attribute

        :param return_data: flag to return or not data downloaded by this request to disk. Default is ``True``.
        :return: bool
        """
        self.return_data = return_data

    def set_data_folder(self, data_folder):
        """
        Set data_folder attribute

        :param data_folder: folder name where the fetched data will be (or already is) saved.
        :return: str
        """
        self.data_folder = data_folder
        self._set_file_location()

    def _set_file_location(self):
        if self.data_folder and self.filename:
            self.file_location = os.path.join(self.data_folder, self.filename.lstrip('/'))
        elif self.filename:
            self.file_location = self.filename

    def is_downloaded(self):
        """ Checks if data for this request has already been downloaded and is saved to disk.

        :return: returns ``True`` if data for this request has already been downloaded and is saved to disk.
        :rtype: bool
        """
        if self.data_folder is None:
            return False
        self._set_file_location()
        return os.path.exists(self.file_location)

    def get_filename(self):
        """ Returns the full filename.

        :return: full filename (data folder + filename)
        :rtype: str
        """
        if self.data_folder:
            file_path = os.path.join(self.data_folder, self.filename)
            if len(file_path) > 255 and sys_is_windows():
                warnings.warn('File path {} is longer than 255 character which might cause an error while saving on'
                              'disk'.format(file_path))
            return file_path
        return None

    def is_aws_s3(self):
        """Checks if data has to be downloaded from AWS s3 bucket

        :return: True if url describes location at AWS s3 bucket and False otherwise
        :rtype: bool
        """
        return self.url.startswith('s3://')


def download_data(request_list, redownload=False, max_threads=None):
    """ Download all requested data or read data from disk, if already downloaded and available and redownload is
    not required.

    :param request_list: list of DownloadRequests
    :type request_list: list of DownloadRequests
    :param redownload: if ``True``, download again the data, although it was already downloaded and is available
                        on the disk. Default is ``False``.
    :type redownload: bool
    :param max_threads: number of threads to use. Default is ``max_threads=None`` (``5*N`` where ``N`` = nr. of cores on
                        the system)
    :type max_threads: int
    :return: list of Futures holding downloaded data, where each element in the list corresponds to an element
                in the download request list.
    :rtype: list[concurrent.futures.Future]
    """
    _check_if_must_download(request_list, redownload)

    LOGGER.debug("Using max_threads=%s for %s requests", max_threads, len(request_list))

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        return [executor.submit(execute_download_request, request) for request in request_list]


def _check_if_must_download(request_list, redownload):
    """
    Updates request.will_download attribute of each request in request_list.

    **Note:** the function mutates the elements of the list!

    :param request_list: a list of ``DownloadRequest`` instances
    :type: list[DownloadRequest]
    :param redownload: tells whether to download the data again or not
    :type: bool
    """
    for request in request_list:
        request.will_download = (request.save_response or request.return_data) \
                                and (not request.is_downloaded() or redownload)


def execute_download_request(request):
    """ Executes download request.

    :param request: DownloadRequest to be executed
    :type request: DownloadRequest
    :return: downloaded data or None
    :rtype: numpy array, other possible data type or None
    :raises: DownloadFailedException
    """
    if request.save_response and request.data_folder is None:
        raise ValueError('Data folder is not specified. '
                         'Please give a data folder name in the initialisation of your request.')

    if not request.will_download:
        return None

    try_num = SHConfig().max_download_attempts
    response = None
    while try_num > 0:
        try:
            if request.is_aws_s3():
                response = _do_aws_request(request)
                response_content = response['Body'].read()
            else:
                response = _do_request(request)
                response.raise_for_status()
                response_content = response.content
            LOGGER.debug('Successful download from %s', request.url)
            break
        except requests.RequestException as exception:
            try_num -= 1
            if try_num > 0 and (_is_temporal_problem(exception) or
                                (isinstance(exception, requests.HTTPError) and
                                 exception.response.status_code >= requests.status_codes.codes.INTERNAL_SERVER_ERROR)):
                LOGGER.debug('Download attempt failed: %s\n%d attempts left, will retry in %ds', exception,
                             try_num, SHConfig().download_sleep_time)
                time.sleep(SHConfig().download_sleep_time)
            else:
                if request.url.startswith(SHConfig().aws_base_url) and isinstance(exception, requests.HTTPError)\
                        and exception.response.status_code == requests.status_codes.codes.NOT_FOUND:
                    raise AwsDownloadFailedException('File in location %s is missing' % request.url)
                raise DownloadFailedException(_create_download_failed_message(exception))

    _save_if_needed(request, response_content)

    if request.return_data:
        return decode_data(response_content, request.data_type, entire_response=response)
    return None


def _do_request(request):
    """ Executes download request
    :param request: A request
    :type request: DownloadRequest
    :return: Response of the request
    :rtype: requests.Response
    """
    if request.request_type is RequestType.GET:
        return requests.get(request.url, headers=request.headers)
    elif request.request_type is RequestType.POST:
        return requests.post(request.url, data=json.dumps(request.post_values), headers=request.headers)
    raise ValueError('Invalid request type {}'.format(request.request_type))


def _do_aws_request(request):
    """ Executes download request from AWS service
    :param request: A request
    :type request: DownloadRequest
    :return: Response of the request
    :rtype: dict
    :raises: AwsDownloadFailedException, ValueError
    """
    if not SHConfig().aws_access_key_id or not SHConfig().aws_secret_access_key:
        raise ValueError('The requested data is in Requester Pays AWS bucket. In order to download the data please set '
                         'your access key in command line:\n'
                         '$ sentinelhub.config --aws_access_key_id <your AWS key> --aws_secret_access_key '
                         '<your AWS secret key>')
    aws_service, _, bucket_name, url_key = request.url.split('/', 3)
    s3_client = boto3.client(aws_service.strip(':'), aws_access_key_id=SHConfig().aws_access_key_id,
                             aws_secret_access_key=SHConfig().aws_secret_access_key)
    try:
        return s3_client.get_object(Bucket=bucket_name, Key=url_key, RequestPayer='requester')
    except s3_client.exceptions.NoSuchKey:
        raise AwsDownloadFailedException('File in location %s is missing' % request.url)
    except s3_client.exceptions.NoSuchBucket:
        raise ValueError('Aws bucket %s does not exist' % bucket_name)


def _is_temporal_problem(exception):
    """ Checks if the obtained exception is temporal and if download attempt should be repeated

    :param exception: Exception raised during download
    :type exception: Exception
    :return: True if exception is temporal and False otherwise
    :rtype: bool
    """
    try:
        return isinstance(exception, (requests.ConnectionError, requests.Timeout))
    except AttributeError:  # Earlier requests versions might not have requests.Timeout
        return isinstance(exception, requests.ConnectionError)


def _create_download_failed_message(exception):
    """ Creates message describing why download has failed

    :param exception: Exception raised during download
    :type exception: Exception
    :return: Error message
    :rtype: str
    """
    message = 'Failed to download with {}:\n{}'.format(exception.__class__.__name__, exception)

    if _is_temporal_problem(exception):
        if isinstance(exception, requests.ConnectionError):
            message += '\nPlease check your internet connection and try again.'
        else:
            message += '\nThere might be a problem in connection or the server failed to process ' \
                       'your request. Please try again.'
    elif isinstance(exception, requests.HTTPError):
        try:
            server_message = ''
            for elem in decode_data(exception.response.content, MimeType.XML):
                if 'ServiceException' in elem.tag or 'Message' in elem.tag:
                    server_message += elem.text.strip('\n\t ')
        except ElementTree.ParseError:
            server_message = exception.response.text
        message += '\nServer response: "{}"'.format(server_message)

    return message


def _save_if_needed(request, response_content):
    """ Save data to disk, if requested by the user
    :param request: Download request
    :type request: DownloadRequest
    :param response_content: content of the download response
    :type response_content: bytes
    """
    if request.save_response:
        create_parent_folder(request.file_location)
        with open(request.file_location, 'wb') as file:
            file.write(response_content)
        LOGGER.debug('Saved data from %s to %s', request.url, request.file_location)


def decode_data(response_content, data_type, entire_response=None):
    """ Interprets downloaded data and returns it.

    :param response_content: downloaded data (i.e. json, png, tiff, xml, zip, ... file)
    :type response_content: bytes
    :param data_type: expected downloaded data type
    :type data_type: constants.MimeType
    :param entire_response: A response obtained from execution of download request
    :type entire_response: requests.Response or dict or None
    :return: downloaded data
    :rtype: numpy array in case of image data type, or other possible data type
    :raises: ValueError
    """
    LOGGER.debug('data_type=%s', data_type)

    if data_type is MimeType.JSON:
        if isinstance(entire_response, requests.Response):
            return entire_response.json()
        return json.loads(response_content.decode('utf-8'))
    elif MimeType.is_image_format(data_type):
        return decode_image(response_content, data_type)
    elif data_type is MimeType.XML or data_type is MimeType.GML or data_type is MimeType.SAFE:
        return ElementTree.fromstring(response_content)
    else:
        return {
            MimeType.RAW: response_content,
            MimeType.TXT: response_content,
            MimeType.REQUESTS_RESPONSE: entire_response,
            MimeType.ZIP: BytesIO(response_content)
        }[data_type]

    raise ValueError('Unknown response data type {}'.format(data_type))


def decode_image(data, image_type):
    """ Decodes the image provided in various formats, i.e. png, 16-bit float tiff, 32-bit float tiff, jp2
        and returns it as an numpy array

    :param data: image in its original format
    :type data: any of possible image types
    :param image_type: expected image format
    :type image_type: constants.MimeType
    :return: image as numpy array
    :rtype: numpy array
    :raises: ImageDecodingError
    """
    if image_type is MimeType.TIFF or image_type is MimeType.TIFF_d32f:
        image = tiff.imread(BytesIO(data))
    else:
        img_array = np.asarray(bytearray(data))
        image = cv2.imdecode(img_array, cv2.IMREAD_UNCHANGED)

        if image_type is MimeType.JP2:
            try:
                bit_depth = get_jp2_bit_depth(BytesIO(data))
                image = _fix_jp2_image(image, bit_depth)
            except ValueError:
                pass

    if image is None:
        raise ImageDecodingError('Unable to decode image')
    return image


def get_json(url, post_values=None, headers=None):
    """ Download request as JSON data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :param post_values: form encoded data to send in POST request. Default is ``None``
    :type post_values: dict
    :param headers: add HTTP headers to request. Default is ``None``
    :type headers: dict
    :return: request response as JSON instance
    :rtype: JSON instance or None
    :raises: RunTimeError
    """

    json_headers = {} if headers is None else headers.copy()

    if post_values is None:
        request_type = RequestType.GET
    else:
        request_type = RequestType.POST
        json_headers = {**json_headers, **{'Content-Type': MimeType.get_string(MimeType.JSON)}}

    request = DownloadRequest(url=url, headers=json_headers, request_type=request_type, post_values=post_values,
                              save_response=False, return_data=True, data_type=MimeType.JSON)

    return execute_download_request(request)


def get_xml(url):
    """ Download request as XML data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :return: request response as XML instance
    :rtype: XML instance or None
    :raises: RunTimeError
    """
    request = DownloadRequest(url=url, request_type=RequestType.GET, save_response=False, return_data=True,
                              data_type=MimeType.XML)

    return execute_download_request(request)



"""
Module for communication with http://opensearch.sentinel-hub.com/resto/api
For more search parameters check: http://opensearch.sentinel-hub.com/resto/api/collections/Sentinel2/describe.xml
"""

import logging
import datetime
from urllib.parse import urlencode

from .constants import CRS
from .config import SHConfig
from .download import get_json
from .geo_utils import transform_bbox


LOGGER = logging.getLogger(__name__)


class TileMissingException(Exception):
    pass


def get_tile_info_id(tile_id):
    """ Get basic information about image tile

    :param tile_id: original tile identification string provided by ESA (e.g.
                    'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :type tile_id: str
    :return: dictionary with info provided by Opensearch REST service or ``None`` if such tile does not exist on AWS.
    :rtype: dict or None
    :raises: TileMissingException if no tile with tile ID `tile_id` exists
    """
    result_list = list(search_iter(tile_id=tile_id))

    if not result_list:
        raise TileMissingException

    if len(result_list) > 1:
        LOGGER.warning('Obtained %d results for tile_id=%d. Returning the first one', len(result_list), tile_id)

    return result_list[0]


def get_tile_info(tile, time, aws_index=None, all_tiles=False):
    """ Get basic information about image tile

    :param tile: tile name (e.g. ``'T10UEV'``)
    :type tile: str
    :param time: time in ISO8601 format
    :type time: str
    :param aws_index: index of tile on AWS
    :type aws_index: int or None
    :param all_tiles: If True it will return list of all tiles otherwise only the first one
    :type all_tiles: bool
    :return: dictionary with info provided by Opensearch REST service or None if such tile does not exist on AWS.
    :rtype: dict or None
    """
    end_date, start_date = _extract_range_from_time(time)

    candidates = []
    for tile_info in search_iter(start_date=start_date, end_date=end_date):
        path_props = tile_info['properties']['s3Path'].split('/')
        this_tile = ''.join(path_props[1:4])
        this_aws_index = int(path_props[-1])
        if this_tile == tile.lstrip('T0') and (aws_index is None or aws_index == this_aws_index):
            candidates.append(tile_info)

    if not candidates:
        raise TileMissingException

    if len(candidates) > 1:
        LOGGER.info('Obtained %d results for tile=%s, time=%s. Returning the first one', len(candidates), tile,
                    time)
    if all_tiles:
        return candidates
    return candidates[0]


def _extract_range_from_time(time):
    """
    Extracts time range from datetime
    :param time: string representation of datetime
    :type: str
    :return: pair of strings of length 2
    :rtype: tuple[str]
    """
    if len(time.split('T')) == 1:
        start_date, end_date = time + 'T00:00:00', time + 'T23:59:59'
    else:
        start_date, end_date = time, time
    return end_date, start_date


def get_area_info(bbox, date_interval, maxcc=None):
    """ Get information about all images from specified area and time range

    :param bbox: bounding box of requested area
    :type bbox: common.BBox
    :param date_interval: a pair of time strings in ISO8601 format
    :type date_interval: tuple(str)
    :param maxcc: filter images by maximum percentage of cloud coverage
    :type maxcc: float in range [0, 1] or None
    :return: list of dictionaries containing info provided by Opensearch REST service
    :rtype: list(dict)
    """
    result_list = search_iter(bbox=bbox, start_date=date_interval[0], end_date=date_interval[1])
    if maxcc:
        return reduce_by_maxcc(result_list, maxcc)
    return result_list


def get_area_dates(bbox, date_interval, maxcc=None):
    """ Get list of times of existing images from specified area and time range

    :param bbox: bounding box of requested area
    :type bbox: common.BBox
    :param date_interval: a pair of time strings in ISO8601 format
    :type date_interval: tuple(str)
    :param maxcc: filter images by maximum percentage of cloud coverage
    :type maxcc: float in range [0, 1] or None
    :return: list of time strings in ISO8601 format
    :rtype: list[datetime.datetime]
    """

    area_info = get_area_info(bbox, date_interval, maxcc=maxcc)
    return sorted(set(
        [datetime.datetime.strptime(tile_info['properties']['startDate'].strip('Z'), '%Y-%m-%dT%H:%M:%S') for tile_info
         in area_info]))


def reduce_by_maxcc(result_list, maxcc):
    """ Filter list image tiles by maximum cloud coverage

    :param result_list: list of dictionaries containing info provided by Opensearch REST service
    :type result_list: list(dict)
    :param maxcc: filter images by maximum percentage of cloud coverage
    :type maxcc: float in range [0, 1]
    :return: list of dictionaries containing info provided by Opensearch REST service
    :rtype: list(dict)
    """
    return [tile_info for tile_info in result_list if tile_info['properties']['cloudCover'] <= 100 * float(maxcc)]


def search_iter(text_query=None, tile_id=None, bbox=None, start_date=None, end_date=None, cloud_cover=None):
    """ Function that implements Opensearch search queries and returns results

    All parameters for search are optional.

    :param text_query: arbitrary text query
    :type text_query: str
    :param tile_id: original tile identification string provided by ESA (e.g.
                    'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :type tile_id: str
    :param bbox: bounding box of requested area
    :type bbox: common.BBox
    :param start_date: beginning of time range in ISO8601 format
    :type start_date: str
    :param end_date: end of time range in ISO8601 format
    :type end_date: str
    :param cloud_cover: percentage of cloud coverage
    :type cloud_cover: float in range [0, 100]
    :return: dictionaries containing info provided by Opensearch REST service
    :rtype: Iterator[dict]
    """
    if bbox and bbox.get_crs() is not CRS.WGS84:
        bbox = transform_bbox(bbox, CRS.WGS84)

    url_params = _prepare_url_params(bbox, cloud_cover, end_date, start_date, text_query, tile_id)
    url_params['maxRecords'] = SHConfig().max_opensearch_records_per_query

    start_index = 1

    while True:
        url_params['index'] = start_index

        url = '{}search.json?{}'.format(SHConfig().opensearch_url, urlencode(url_params))
        LOGGER.debug("URL=%s", url)

        response = get_json(url)
        for tile_info in response["features"]:
            yield tile_info

        if len(response["features"]) < SHConfig().max_opensearch_records_per_query:
            break
        start_index += SHConfig().max_opensearch_records_per_query


def _prepare_url_params(bbox, cloud_cover, end_date, start_date, text_query, tile_id):
    """ Constructs dict with URL params

    :param bbox: bounding box of requested area in WGS84 CRS
    :type bbox: common.BBox
    :param cloud_cover: percentage of cloud coverage
    :type cloud_cover: float in range [0, 100]
    :param start_date: beginning of time range in ISO8601 format
    :type start_date: str
    :param end_date: end of time range in ISO8601 format
    :type end_date: str
    :param text_query: arbitrary text query
    :type text_query: str
    :param tile_id: original tile identification string provided by ESA (e.g.
                    'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :type tile_id: str
    :return: dictionary with parameters as properties when arguments not None
    :rtype: dict
    """
    url_params = _add_param({}, text_query, 'q')
    url_params = _add_param(url_params, tile_id, 'identifier')
    url_params = _add_param(url_params, start_date, 'startDate')
    url_params = _add_param(url_params, end_date, 'completionDate')
    url_params = _add_param(url_params, cloud_cover, 'cloudCover')
    if bbox:
        url_params = _add_param(url_params, str(bbox), 'box')
    return url_params


def _add_param(params, value, key):
    """ If value is not None then return dict params with added (key, value) pair

    :param params: dictionary of parameters
    :type: dict
    :param value: Value
    :param key: Key
    :return: if value not ``None`` then a copy of params with (key, value) added, otherwise returns params
    """
    if value:
        params[key] = value
    return params



"""Module implementing classes common to all modules of the package (such as the bounding box class).

Representing bounding box can be confusing. If a function expects bbox as a list of four coordinates,
is it [lng1, lat1, lng2, lat2]? Or is it something else? And what CRS does it expect? Is the CRS a
separate parameter?

In this module the BBox class provides the canonical representation of a BBox that all the functions and
classes of the sg_utils package use, solving these issues.

Available classes:
 - BBox, represent a bounding box in a given CRS
"""

from .constants import CRS


class BBox:
    """ Class representing a bounding box in a given CRS.

    Throughout the sentinelhub package this class serves as the canonical representation of a bounding
    box. It can instantiate itself from multiple representations:

        1) ``((min_x,min_y),(max_x,max_y))``,
        2) ``(min_x,min_y,max_x,max_y)``,
        3) ``[min_x,min_y,max_x,max_y]``,
        4) ``[[min_x, min_y],[max_x,max_y]]``,
        5) ``[(min_x, min_y),(max_x,max_y)]``,
        6) ``([min_x, min_y],[max_x,max_y])``,
        7) ``'min_x,min_y,max_x,max_y'``,
        8) ``{'min_x':min_x, 'max_x':max_x, 'min_y':min_y, 'max_y':max_y}``,
        9) ``bbox``, where ``bbox`` is an instance of ``BBox``.

    Note that BBox coordinate system depends on ``crs`` parameter:
        - In case of ``constants.CRS.WGS84`` axis x represents longitude and axis y represents latitude
        - In case of ``constants.CRS.POP_WEB`` axis x represents easting and axis y represents northing
        - In case of ``constants.CRS.UTM_*`` axis x represents easting and axis y represents northing

    :param bbox: a bbox in a number of representations.
    :param crs: Coordinate Reference System that bbox is in. Expect one of the constants from the ``const.CRS`` enum.
    :type crs: constants.CRS
    """
    def __init__(self, bbox, crs):
        x_fst, y_fst, x_snd, y_snd = BBox._to_tuple(bbox)
        self.min_x = min(x_fst, x_snd)
        self.max_x = max(x_fst, x_snd)
        self.min_y = min(y_fst, y_snd)
        self.max_y = max(y_fst, y_snd)
        self.crs = CRS(crs)

    def __iter__(self):
        return iter(self.get_lower_left() + self.get_upper_right())

    def get_lower_left(self):
        """ Returns the lower left vertex of the bounding box

        :return: min_x, min_y
        :rtype: (float, float)
        """
        return self.min_x, self.min_y

    def get_upper_right(self):
        """ Returns the upper right vertex of the bounding box

        :return: max_x, max_y
        :rtype: (float, float)
        """
        return self.max_x, self.max_y

    def get_middle(self):
        """ Returns the middle point of the bounding box

        :return: middle point
        :rtype: (float, float)
        """
        return (self.min_x + self.max_x) / 2, (self.min_y + self.max_y) / 2

    def get_crs(self):
        """ Returns the coordinate reference system (CRS) of the bounding box.

        :return: CRS that the BBox is given in
        :rtype: constants.CRS
        """
        return self.crs

    def get_polygon(self, reverse=False):
        """ Returns a list of coordinates of 5 points describing a polygon. Points are listed in clockwise order, first
        point is the same as the last.

        :param reverse: True if x and y coordinates should be switched and False otherwise
        :type reverse: bool
        :return: [[x_1, y_1], ... , [x_5, y_5]]
        :rtype: list(list(float))
        """
        polygon = [[self.min_x, self.min_y],
                   [self.min_x, self.max_y],
                   [self.max_x, self.max_y],
                   [self.max_x, self.min_y],
                   [self.min_x, self.min_y]]
        if reverse:
            for i, point in enumerate(polygon):
                polygon[i] = point[::-1]
        return polygon

    def get_partition(self, num_x=1, num_y=1):
        """ Partitions bounding box into smaller bounding boxes of the same size.

        :param num_x: Number of parts BBox will be horizontally divided into.
        :type num_x: int
        :param num_y: Number of parts BBox will be vertically divided into.
        :type num_y: int or None
        :return: Two-dimensional list of smaller bounding boxes. Their location is
        :rtype: list(list(BBox))
        """
        size_x, size_y = (self.max_x - self.min_x) / num_x, (self.max_y - self.min_y) / num_y
        return [[BBox([self.min_x + i * size_x, self.min_y + j * size_y,
                       self.min_x + (i + 1) * size_x, self.min_y + (j + 1) * size_y],
                      crs=self.crs) for j in range(num_y)] for i in range(num_x)]

    def __repr__(self):
        return "{}((({}, {}), ({}, {})), crs={})".format(self.__class__.__name__, self.min_x, self.min_y, self.max_x,
                                                         self.max_y, self.crs)

    def __str__(self, reverse=False):
        """ Transforms bounding box into string of coordinates

        :param reverse: True if x and y coordinates should be switched and False otherwise
        :type reverse: bool
        :return: String of coordinates
        :rtype: str
        """
        if reverse:
            return "{},{},{},{}".format(self.min_y, self.min_x, self.max_y, self.max_x)
        return "{},{},{},{}".format(self.min_x, self.min_y, self.max_x, self.max_y)

    @staticmethod
    def _to_tuple(bbox):
        """ Converts the input bbox representation (see the constructor docstring for a list of valid representations)
            into a flat tuple

        :param bbox: A bbox in one of 7 forms listed in the class description.
        :return: A flat tuple of size
        :raises: TypeError
        """
        if isinstance(bbox, (list, tuple)):
            return BBox._tuple_from_list_or_tuple(bbox)
        elif isinstance(bbox, str):
            return BBox._tuple_from_str(bbox)
        elif isinstance(bbox, dict):
            return BBox._tuple_from_dict(bbox)
        elif isinstance(bbox, BBox):
            return BBox._tuple_from_bbox(bbox)
        raise TypeError('Invalid bbox representation')

    @staticmethod
    def _tuple_from_list_or_tuple(bbox):
        """ Converts a list or tuple representation of a bbox into a flat tuple representation.

        :param bbox: a list or tuple with 4 coordinates that is either flat or nested
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: TypeError
        """
        if len(bbox) == 4:
            return tuple(map(float, bbox))
        elif len(bbox) == 2 and all([isinstance(point, (list, tuple)) for point in bbox]):
            return BBox._tuple_from_list_or_tuple(bbox[0] + bbox[1])
        raise TypeError('Expected a valid list or tuple representation of a bbox')

    @staticmethod
    def _tuple_from_str(bbox):
        """ Parses a string of numbers separated by any combination of commas and spaces

        :param bbox: e.g. str of the form 'min_x ,min_y  max_x, max_y'
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return tuple([float(s) for s in bbox.replace(',', ' ').split() if s])

    @staticmethod
    def _tuple_from_dict(bbox):
        """ Converts a dictionary representation of a bbox into a flat tuple representation

        :param bbox: a dict with keys "min_x, "min_y", "max_x", and "max_y"
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: KeyError
        """
        return bbox['min_x'], bbox['min_y'], bbox['max_x'], bbox['max_y']

    @staticmethod
    def _tuple_from_bbox(bbox):
        """ Converts a BBox instance into a tuple

        :param bbox: An instance of the BBox type
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return bbox.get_lower_left() + bbox.get_upper_right()



"""
Module for creating .SAFE structure with data from AWS
"""

import warnings

from .download import get_xml
from .constants import AwsConstants, EsaSafeType, MimeType, DataSource
from .aws import AwsProduct, AwsTile


class SafeProduct(AwsProduct):
    """ Class inherits from `aws.AwsProduct`"""
    def get_requests(self):
        """
        Creates product structure and returns list of files for download

        :return: list of download requests
        :rtype: list(download.DownloadRequest)
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.parent_folder)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_safe_struct(self):
        """
        Describes a structure inside tile folder of ESA product .SAFE structure

        :return: nested dictionaries representing .SAFE structure
        :rtype: dict
        """
        safe = {}
        main_folder = self.get_main_folder()
        safe[main_folder] = {}

        safe[main_folder][AwsConstants.AUX_DATA] = {}

        safe[main_folder][AwsConstants.DATASTRIP] = {}
        datastrip_list = self.get_datastrip_list()
        for datastrip_folder, datastrip_url in datastrip_list:
            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder] = {}
            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder][AwsConstants.QI_DATA] = {}
            # S-2 L1C reports are on AWS only stored with tiles and without RADIOMETRIC_QUALITY
            if self.has_reports() and self.data_source is DataSource.SENTINEL2_L2A:
                for metafile in AwsConstants.QUALITY_REPORTS:
                    metafile_name = self.add_file_extension(metafile)
                    safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder][AwsConstants.QI_DATA][
                        metafile_name] = '{}/qi/{}'.format(datastrip_url, metafile_name)

            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder][
                self.get_datastrip_metadata_name(datastrip_folder)] = '{}/{}'.format(
                    datastrip_url, self.add_file_extension(AwsConstants.METADATA))

        safe[main_folder][AwsConstants.GRANULE] = {}

        for tile_info in self.product_info['tiles']:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_struct = SafeTile(tile_name, date, aws_index, parent_folder=None, bands=self.bands,
                                       metafiles=self.metafiles, data_source=self.data_source).get_safe_struct()
                for tile_name, safe_struct in tile_struct.items():
                    safe[main_folder][AwsConstants.GRANULE][tile_name] = safe_struct

        safe[main_folder][AwsConstants.HTML] = {}  # AWS doesn't have this data
        safe[main_folder][AwsConstants.INFO] = {}  # AWS doesn't have this data

        safe[main_folder][self.get_product_metadata_name()] = self.get_url(AwsConstants.METADATA)
        safe[main_folder]['INSPIRE.xml'] = self.get_url(AwsConstants.INSPIRE)
        safe[main_folder][self.add_file_extension(AwsConstants.MANIFEST)] = self.get_url(AwsConstants.MANIFEST)

        if self.is_early_compact_l2a():
            safe[main_folder]['L2A_Manifest.xml'] = self.get_url(AwsConstants.L2A_MANIFEST)
            safe[main_folder][self.get_report_name()] = self.get_url(AwsConstants.REPORT)

        if self.safe_type == EsaSafeType.OLD_TYPE and self.baseline != '02.02':
            safe[main_folder][_edit_name(self.product_id, 'BWI') + '.png'] = self.get_url(AwsConstants.PREVIEW,
                                                                                          MimeType.PNG)
        return safe

    def get_main_folder(self):
        """
        :return: name of main folder
        :rtype: str
        """
        return '{}.SAFE'.format(self.product_id)

    def get_datastrip_list(self):
        """
        :return: list of datastrips folder names and urls from productInfo.json file
        :rtype: list((str, str))
        """
        datastrips = self.product_info['datastrips']
        return [(self.get_datastrip_name(datastrip['id']), self.base_url + datastrip['path'])
                for datastrip in datastrips]

    def get_datastrip_name(self, datastrip):
        """
        :param datastrip: name of datastrip
        :type datastrip: str
        :return: name of datastrip folder
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            return datastrip
        return '_'.join(datastrip.split('_')[4:-1])

    def get_datastrip_metadata_name(self, datastrip_folder):
        """
        :param datastrip_folder: name of datastrip folder
        :type: str
        :return: name of datastrip metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = _edit_name(datastrip_folder, 'MTD', delete_end=True)
        else:
            name = 'MTD_DS'
        return '{}.{}'.format(name, MimeType.XML.value)

    def get_product_metadata_name(self):
        """
        :return: name of product metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = _edit_name(self.product_id, 'MTD', 'SAFL1C')
        else:
            name = 'MTD_{}'.format(self.product_id.split('_')[1])
        return '{}.{}'.format(name, MimeType.XML.value)

    def get_report_name(self):
        """
        :return: name of the report file of L2A products
        :rtype: str
        """
        return '{}_{}_report.{}'.format(self.product_id, self.get_report_time(), MimeType.XML.value)

    def get_report_time(self):
        """ Returns time when the L2A processing started and reports was created.
        :return: String in a form YYYYMMDDTHHMMSS
        :rtype: str
        """
        tree = get_xml(self.get_url(AwsConstants.REPORT))

        try:
            timestamp = tree.find('check/inspection').attrib['execution']
            return timestamp.split(',')[0].replace(' ', 'T').replace(':', '').replace('-', '')
        except AttributeError:
            warnings.warn('Could not obtain the L2A report creation time')
            return 'unknown'


class SafeTile(AwsTile):
    """ Class inherits from `aws.AwsTile`"""
    def __init__(self, *args, **kwargs):
        super(SafeTile, self).__init__(*args, **kwargs)

        self.tile_id = self.get_tile_id()

    def get_requests(self):
        """
        Creates tile structure and returns list of files for download.

        :return: list of download requests for
        :rtype: list(download.DownloadRequest)
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.parent_folder)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_safe_struct(self):
        """
        Describes a structure inside tile folder of ESA product .SAFE structure.

        :return: nested dictionaries representing .SAFE structure
        :rtype: dict
        """
        # pylint: disable=too-many-branches
        safe = {}
        main_folder = self.get_main_folder()
        safe[main_folder] = {}

        safe[main_folder][AwsConstants.AUX_DATA] = {}
        # Not sure if 2nd condition of the following is correct:
        if self.data_source is not DataSource.SENTINEL2_L1C or self.baseline != '02.04':
            ecmwft_file = AwsConstants.ECMWFT if self.data_source is DataSource.SENTINEL2_L1C or \
                                                 self.safe_type is EsaSafeType.OLD_TYPE else AwsConstants.AUX_ECMWFT
            safe[main_folder][AwsConstants.AUX_DATA][self.get_aux_data_name()] = self.get_url(ecmwft_file)
        # Old products also have DEM and MSI in aux folder

        if self.is_early_compact_l2a():
            safe[main_folder][AwsConstants.AUX_DATA][self.add_file_extension(AwsConstants.GIPP, remove_path=True)] =\
                self.get_url(AwsConstants.GIPP)

        safe[main_folder][AwsConstants.IMG_DATA] = {}
        if self.data_source is DataSource.SENTINEL2_L1C:
            for band in self.bands:
                safe[main_folder][AwsConstants.IMG_DATA][self.get_img_name(band)] = self.get_url(band)
            if self.safe_type == EsaSafeType.COMPACT_TYPE:
                safe[main_folder][AwsConstants.IMG_DATA][self.get_img_name(AwsConstants.TCI)] =\
                    self.get_url(AwsConstants.TCI)
        else:
            for resolution in AwsConstants.RESOLUTIONS:
                safe[main_folder][AwsConstants.IMG_DATA][resolution] = {}
            for band_name in self.bands:
                resolution, band = band_name.split('/')
                if self._band_exists(band_name):
                    safe[main_folder][AwsConstants.IMG_DATA][resolution][self.get_img_name(band, resolution)] =\
                        self.get_url(band_name)

        safe[main_folder][AwsConstants.QI_DATA] = {}
        safe[main_folder][AwsConstants.QI_DATA][self.get_qi_name('CLOUDS')] = self.get_gml_url('CLOUDS')
        for qi_type in AwsConstants.QI_LIST:
            for band in AwsConstants.S2_L1C_BANDS:
                safe[main_folder][AwsConstants.QI_DATA][self.get_qi_name(qi_type, band)] = self.get_gml_url(qi_type,
                                                                                                            band)

        if self.has_reports():
            for metafile in [AwsConstants.FORMAT_CORRECTNESS, AwsConstants.GENERAL_QUALITY,
                             AwsConstants.GEOMETRIC_QUALITY, AwsConstants.SENSOR_QUALITY]:
                metafile_name = self.add_file_extension(metafile, remove_path=True)
                safe[main_folder][AwsConstants.QI_DATA][metafile_name] = self.get_qi_url(metafile_name)

        if self.data_source is DataSource.SENTINEL2_L2A:
            for mask in AwsConstants.CLASS_MASKS:
                for resolution in [AwsConstants.R20m, AwsConstants.R60m]:
                    if self.baseline <= '02.06':
                        mask_name = self.get_img_name(mask, resolution)
                    else:
                        mask_name = self.get_qi_name('{}PRB'.format(mask), resolution.lstrip('R'), MimeType.JP2)
                    safe[main_folder][AwsConstants.QI_DATA][mask_name] =\
                        self.get_qi_url('{}_{}.jp2'.format(mask, resolution.lstrip('R')))

        if self.is_early_compact_l2a():
            safe[main_folder][AwsConstants.QI_DATA][self.get_img_name(AwsConstants.PVI)] = self.get_preview_url('L2A')

        preview_type = 'L2A' if self.data_source is DataSource.SENTINEL2_L2A and self.baseline >= '02.07' else 'L1C'
        safe[main_folder][AwsConstants.QI_DATA][self.get_preview_name()] = self.get_preview_url(preview_type)

        safe[main_folder][self.get_tile_metadata_name()] = self.get_url(AwsConstants.METADATA)

        return safe

    def get_tile_id(self):
        """Creates ESA tile ID

        :return: ESA tile ID
        :rtype: str
        """
        tree = get_xml(self.get_url(AwsConstants.METADATA))

        tile_id_tag = 'TILE_ID_2A' if self.data_source is DataSource.SENTINEL2_L2A and self.baseline <= '02.06' else\
            'TILE_ID'
        tile_id = tree[0].find(tile_id_tag).text
        if self.safe_type is not EsaSafeType.OLD_TYPE:
            info = tile_id.split('_')
            tile_id = '_'.join([info[3], info[-2], info[-3], self.get_sensing_time()])
        return tile_id

    def get_sensing_time(self):
        """
        :return: Exact tile sensing time
        :rtype: str
        """
        return self.tile_info['timestamp'].split('.')[0].replace('-', '').replace(':', '')

    def get_datatake_time(self):
        """
        :return: Exact time of datatake
        :rtype: str
        """
        return self.tile_info['productName'].split('_')[2]

    def get_main_folder(self):
        """
        :return: name of tile folder
        :rtype: str
        """
        return self.tile_id

    def get_tile_metadata_name(self):
        """
        :return: name of tile metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, 'MTD', delete_end=True)
        else:
            name = 'MTD_TL'
        return '{}.xml'.format(name)

    def get_aux_data_name(self):
        """
        :return: name of auxiliary data file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            # this is not correct, but we cannot reconstruct last two timestamps in auxiliary data file name
            # e.g. S2A_OPER_AUX_ECMWFT_EPA__20160120T231011_V20160103T150000_20160104T030000
            return 'AUX_ECMWFT'
        return 'AUX_ECMWFT'

    def get_img_name(self, band, resolution=None):
        """
        :param band: band name
        :type band: str
        :param resolution: Specifies the resolution in case of Sentinel-2 L2A products
        :type resolution: str or None
        :return: name of band image file
        :rtype: str
        """
        band = band.split('/')[-1]
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = self.tile_id.rsplit('_', 1)[0] + '_' + band
        else:
            name = '_'.join([self.tile_id.split('_')[1], self.get_datatake_time(), band])
        if self.data_source is DataSource.SENTINEL2_L2A and resolution is not None:
            name = '{}_{}'.format(name, resolution.lstrip('R'))
        if self.data_source is DataSource.SENTINEL2_L2A and self.baseline <= '02.06':
            name = 'L2A_{}'.format(name)
        return '{}.jp2'.format(name)

    def get_qi_name(self, qi_type, band='B00', data_format=MimeType.GML):
        """
        :param qi_type: type of quality indicator
        :type qi_type: str
        :param band: band name
        :type band: str
        :param data_format: format of the file
        :type data_format: MimeType
        :return: name of gml file
        :rtype: str
        """
        band = band.split('/')[-1]
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, 'MSK', delete_end=True)
            source_param = '{}_TL'.format('L1C' if self.data_source is DataSource.SENTINEL2_L1C else 'L2A')
            name = name.replace(source_param, qi_type)
            name = '{}_{}_MSIL1C'.format(name, band)
        else:
            name = 'MSK_{}_{}'.format(qi_type, band)
        return '{}.{}'.format(name, data_format.value)

    def get_preview_name(self):
        """Returns .SAFE name of full resolution L1C preview
        :return: name of preview file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, AwsConstants.PVI, delete_end=True)
        else:
            name = '_'.join([self.tile_id.split('_')[1], self.get_datatake_time(), AwsConstants.PVI])
        return '{}.jp2'.format(name)


def _edit_name(name, code, add_code=None, delete_end=False):
    """
    Helping function for creating file names in .SAFE format

    :param name: initial string
    :type name: str
    :param code:
    :type code: str
    :param add_code:
    :type add_code: str or None
    :param delete_end:
    :type delete_end: bool
    :return: edited string
    :rtype: str
    """
    info = name.split('_')
    info[2] = code
    if add_code is not None:
        info[3] = add_code
    if delete_end:
        info.pop()
    return '_'.join(info)



"""
Module with enum constants and utm utils
"""

# pylint: disable=invalid-name

import itertools as it
import mimetypes
import utm
import os.path
from pyproj import Proj
from enum import Enum, EnumMeta


mimetypes.add_type('application/json', '.json')


class PackageProps:
    """ Class for obtaining package properties. Currently it supports obtaining package version."""

    @staticmethod
    def get_version():
        for line in open(os.path.join(os.path.dirname(__file__), '__init__.py')):
            if line.find("__version__") >= 0:
                version = line.split("=")[1].strip()
                version = version.strip('"').strip("'")
        return version


class ServiceType(Enum):
    """ Enum constant class for type of service

    Supported types are WMS, WCS, WFS, AWS
    """
    WMS = 'wms'
    WCS = 'wcs'
    WFS = 'wfs'
    AWS = 'aws'


class _DataSourceMeta(EnumMeta):
    """ EnumMeta class for `DataSource` Enum class
    """
    def __iter__(cls):
        return (member for name, member in cls._member_map_.items() if isinstance(member.value, tuple))

    def __len__(cls):
        return len(list(cls.__iter__()))


class DataSource(Enum, metaclass=_DataSourceMeta):
    """ Enum constant class for types of satellite data

    Supported types are SENTINEL2_L1C, SENTINEL2_L2A, LANDSAT8, SENTINEL1_IW, SENTINEL1_EW, SENTINEL1_EW_SH, DEM, MODIS
    """
    class Source(Enum):
        """
        Types of satellite sources
        """
        SENTINEL2 = 'Sentinel-2'
        SENTINEL1 = 'Sentinel-1'
        LANDSAT8 = 'Landsat 8'
        MODIS = 'MODIS'
        DEM = 'Mapzen DEM'

    class ProcessingLevel(Enum):
        """
        Types of processing level
        """
        L1C = 'L1C'
        L2A = 'L2A'
        GRD = 'GRD'
        MCD43A4 = 'MCD43A4'

    class Acquisition(Enum):
        """
        Types of Sentinel-1 acquisition
        """
        IW = 'IW'
        EW = 'EW'

    class Polarisation(Enum):
        """
        Types of Sentinel-1 polarisation
        """
        DV = 'VV+VH'
        DH = 'HH+HV'
        SV = 'VV'
        SH = 'HH'

    class Resolution(Enum):
        """
        Types of Sentinel-1 resolution
        """
        MEDIUM = 'medium'
        HIGH = 'high'

    class OrbitDirection(Enum):
        """
        Types of Sentinel-1 orbit direction
        """
        ASCENDING = 'ascending'
        DESCENDING = 'descending'
        BOTH = 'both'

    SENTINEL2_L1C = (Source.SENTINEL2, ProcessingLevel.L1C)
    SENTINEL2_L2A = (Source.SENTINEL2, ProcessingLevel.L2A)
    SENTINEL1_IW = (Source.SENTINEL1, ProcessingLevel.GRD, Acquisition.IW, Polarisation.DV, Resolution.HIGH,
                    OrbitDirection.BOTH)
    SENTINEL1_EW = (Source.SENTINEL1, ProcessingLevel.GRD, Acquisition.EW, Polarisation.DH, Resolution.MEDIUM,
                    OrbitDirection.BOTH)
    SENTINEL1_EW_SH = (Source.SENTINEL1, ProcessingLevel.GRD, Acquisition.EW, Polarisation.SH, Resolution.MEDIUM,
                       OrbitDirection.BOTH)
    DEM = (Source.DEM, )
    MODIS = (Source.MODIS, ProcessingLevel.MCD43A4)
    LANDSAT8 = (Source.LANDSAT8, ProcessingLevel.GRD)

    @classmethod
    def get_wfs_typename(cls, data_source):
        """ Maps data source to string identifier for WFS

        :param data_source: One of the supported data sources
        :type: DataSource
        :return: Product identifier for WFS
        :rtype: str
        """
        return {
            cls.SENTINEL2_L1C: 'S2.TILE',
            cls.SENTINEL2_L2A: 'DSS2',
            cls.SENTINEL1_IW: 'DSS3',
            cls.SENTINEL1_EW: 'DSS3',
            cls.SENTINEL1_EW_SH: 'DSS3',
            cls.DEM: 'DSS4',
            cls.MODIS: 'DSS5',
            cls.LANDSAT8: 'DSS6'
        }[data_source]

    @classmethod
    def is_sentinel1(cls, data_source):
        """Checks if source is Sentinel-1

        :param data_source: One of the supported data sources
        :type: DataSource
        :return: ``True`` if source is Sentinel-1 and ``False`` otherwise
        :rtype: bool
        """
        return data_source.value[0] is cls.Source.value.SENTINEL1

    @classmethod
    def is_timeless(cls, data_source):
        """Checks if data source is time independent

        :param data_source: One of the supported data sources
        :type: DataSource
        :return: ``True`` if data source is time independent and ``False`` otherwise
        :rtype: bool
        """
        return data_source.value[0] is cls.Source.value.DEM

    @classmethod
    def is_uswest_source(cls, data_source):
        """Checks if data source via Sentinel Hub services is available at US West server

        :param data_source: One of the supported data sources
        :type: DataSource
        :return: ``True`` if data source exists at US West server and ``False`` otherwise
        :rtype: bool
        """
        return data_source.value[0] in [cls.Source.value.LANDSAT8, cls.Source.value.MODIS, cls.Source.value.DEM]


class _Direction(Enum):
    """ Enum constant class to encode NORTH/SOUTH direction """
    NORTH = 'N'
    SOUTH = 'S'


def _get_utm_code(zone, direction):
    """ Get UTM code given a zone and direction

    Direction is encoded as NORTH=6, SOUTH=7, while zone is the UTM zone number zero-padded.
    For instance, the code 32604 is returned for zone number 4, north direction.

    :param zone: UTM zone number
    :type zone: int
    :param direction: Direction enum type
    :type direction: Enum
    :return: UTM code
    :rtype: str
    """
    dir_dict = {_Direction.NORTH: '6', _Direction.SOUTH: '7'}
    return '{}{}{}'.format('32', dir_dict[direction], str(zone).zfill(2))


def _get_utm_name_value_pair(zone, direction=_Direction.NORTH):
    """ Get name and code for UTM coordinates

    :param zone: UTM zone number
    :type zone: int
    :param direction: Direction enum type
    :type direction: Enum, optional (default=NORTH)
    :return: Name and code of UTM coordinates
    :rtype: str, str
    """
    name = 'UTM_{}{}'.format(zone, direction.value)
    epsg = _get_utm_code(zone, direction)
    return name, epsg


class _BaseCRS(Enum):
    """ Coordinate Reference System enumerate class """
    def __str__(self):
        return self.ogc_string(self)

    @staticmethod
    def get_utm_from_wgs84(lng, lat):
        """ Convert from WGS84 to UTM coordinate system

        :param lng: Longitude
        :type lng: float
        :param lat: Latitude
        :type lat: float
        :return: UTM coordinates
        :rtype: tuple
        """
        _, _, zone, _ = utm.from_latlon(lat, lng)
        direction = 'N' if lat >= 0 else 'S'
        return CRS['UTM_{}{}'.format(str(zone), direction)]

    @staticmethod
    def ogc_string(crs):
        """ Returns a string of the form authority:id representing the CRS.

        :param crs: An enum constant representing a coordinate reference system.
        :type crs: Enum constant
        :return: A string representation of the CRS.
        :rtype: str
        """
        return 'EPSG:' + CRS(crs).value

    @staticmethod
    def is_utm(crs):
        """ Checks if crs is one of the 64 possible UTM coordinate reference systems.

        :param crs: An enum constant representing a coordinate reference system.
        :type crs: Enum constant
        :return: True if crs is UTM and False otherwise
        :rtype: bool
        """
        return crs.name.startswith('UTM')

    @classmethod
    def projection(cls, crs):
        """ Returns a projection in form of pyproj class

        :param crs: An enum constant representing a coordinate reference system.
        :type crs: Enum constant
        :return: pyproj projection class
        :rtype: pyproj.Proj
        """
        return Proj(init=cls.ogc_string(crs))

    @classmethod
    def has_value(cls, value):
        """ Tests whether CRS contains a constant defined with string ``value``.

        :param value: The string representation of the enum constant.
        :type value: str
        :return: ``True`` if there exists a constant with string value ``value``, ``False`` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)


# Look-up class with possible combinations of UTM zone and direction
CRS = _BaseCRS("CRS", dict(
    [_get_utm_name_value_pair(zone, direction) for zone, direction in it.product(range(1, 65), _Direction)] +
    [('WGS84', '4326'), ('POP_WEB', '3857')]
))


def _crs_parser(cls, value):
    """ Parses user input for class CRS

    :param cls: class object
    :param value: user input for CRS
    :type value: str, int or CRS
    """
    parsed_value = value
    if isinstance(parsed_value, int):
        parsed_value = str(parsed_value)
    if isinstance(parsed_value, str):
        parsed_value = parsed_value.strip('epsgEPSG: ')
    return super(_BaseCRS, cls).__new__(cls, parsed_value)


setattr(CRS, '__new__', _crs_parser)


class CustomUrlParam(Enum):
    """ Enum class to represent supported custom url parameters of OGC services

    Supported parameters are `ShowLogo`, `AtmFilter`, `EvalScript`, `EvalScriptUrl`, `Preview`, `Quality`, `Upsampling`,
    `Downsampling`, `Transparent`, `BgColor` and `Geometry`.

    See http://sentinel-hub.com/develop/documentation/api/custom-url-parameters and
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wms-parameters for more information.
    """
    SHOWLOGO = 'ShowLogo'
    ATMFILTER = 'AtmFilter'
    EVALSCRIPT = 'EvalScript'
    EVALSCRIPTURL = 'EvalScriptUrl'
    PREVIEW = 'Preview'
    QUALITY = 'Quality'
    UPSAMPLING = 'Upsampling'
    DOWNSAMPLING = 'Downsampling'
    TRANSPARENT = 'Transparent'
    BGCOLOR = 'BgColor'
    GEOMETRY = 'Geometry'

    @classmethod
    def has_value(cls, value):
        """ Tests whether CustomUrlParam contains a constant defined with a string `value`

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with a string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value.lower() == item.value.lower() for item in cls)

    @staticmethod
    def get_string(param):
        """ Get custom url parameter name as string

        :param param: CustomUrlParam enum constant
        :type param: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        return param.value


class MimeType(Enum):
    """ Enum class to represent supported image file formats

    Supported file formats are TIFF 8-bit, TIFF 16-bit, TIFF 32-bit float, PNG, JPEG, JPEG2000, JSON, CSV, ZIP, HDF5,
    XML, GML, RAW
    """
    TIFF = 'tiff'
    TIFF_d8 = 'tiff;depth=8'
    TIFF_d16 = 'tiff;depth=16'  # This is the same as TIFF
    TIFF_d32f = 'tiff;depth=32f'
    PNG = 'png'
    JPG = 'jpg'
    JP2 = 'jp2'
    JSON = 'json'
    CSV = 'csv'
    ZIP = 'zip'
    HDF = 'hdf'
    XML = 'xml'
    GML = 'gml'
    TXT = 'txt'
    RAW = 'raw'
    SAFE = 'safe'
    REQUESTS_RESPONSE = 'response'  # http://docs.python-requests.org/en/master/api/#requests.Response

    @staticmethod
    def canonical_extension(fmt_ext):
        """ Canonical extension of file format extension

        Converts the format extension fmt_ext into the canonical extension for that format. For example,
        ``canonical_extension('tif') == 'tiff'``. Here we agree that the canonical extension for format F is F.value

        :param fmt_ext: A string representing an extension (e.g. ``'txt'``, ``'png'``, etc.)
        :type fmt_ext: str
        :return: The canonical form of the extension (e.g. if ``fmt_ext='tif'`` then we return ``'tiff'``)
        :rtype: str
        """
        if MimeType.has_value(fmt_ext):
            return fmt_ext
        try:
            return {
                'tif': MimeType.TIFF.value,
                'jpeg': MimeType.JPG.value,
                'hdf5': MimeType.HDF.value,
                'h5': MimeType.HDF.value
            }[fmt_ext]
        except KeyError:
            raise ValueError('Data format .{} is not supported'.format(fmt_ext))

    @staticmethod
    def is_image_format(value):
        """ Checks whether file format is an image format

        :param value: File format
        :type value: str
        :return: ``True`` if file is in image format, ``False`` otherwise
        :rtype: bool
        """
        return value in frozenset([MimeType.TIFF, MimeType.TIFF_d8, MimeType.TIFF_d16, MimeType.TIFF_d32f, MimeType.PNG,
                                   MimeType.JP2, MimeType.JPG])

    @classmethod
    def has_value(cls, value):
        """ Tests whether MimeType contains a constant defined with string ``value``

        :param value: The string representation of the enum constant
        :type value: str
        :return: ``True`` if there exists a constant with string value ``value``, ``False`` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)

    @staticmethod
    def get_string(fmt):
        """ Get file format as string

        :param fmt: MimeType enum constant
        :type fmt: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        if fmt in [MimeType.TIFF_d8, MimeType.TIFF_d16, MimeType.TIFF_d32f]:
            return 'image/{}'.format(fmt.value)
        elif fmt is MimeType.JP2:
            return 'image/jpeg2000'
        elif fmt in [MimeType.RAW, MimeType.REQUESTS_RESPONSE]:
            return fmt.value
        return mimetypes.types_map['.' + fmt.value]


class RequestType(Enum):
    """ Enum constant class for GET/POST request type """
    GET = 'GET'
    POST = 'POST'


class OgcConstants:
    """ Initialisation of constants used by OGC request.

        Constants are LATEST
    """
    LATEST = 'latest'
    HEADERS = {'User-Agent': 'sentinelhub-py/v{}'.format(PackageProps.get_version())}


class AwsConstants:
    """ Initialisation of every constant used by AWS classes

    For each supported data source it contains lists of all possible bands and all posible metadata files:

        - S2_L1C_BANDS and S2_L1C_METAFILES
        - S2_L2A_BANDS and S2_L2A_METAFILES

    It also contains dictionary of all possible files and their formats: AWS_FILES
    """
    # General constants:
    SOURCE_ID_LIST = ['L1C', 'L2A']
    TILE_INFO = 'tileInfo'
    PRODUCT_INFO = 'productInfo'
    METADATA = 'metadata'
    PREVIEW = 'preview'
    PREVIEW_JP2 = 'preview*'
    QI_LIST = ['DEFECT', 'DETFOO', 'NODATA', 'SATURA', 'TECQUA']
    QI_MSK_CLOUD = 'qi/MSK_CLOUDS_B00'
    AUX_DATA = 'AUX_DATA'
    DATASTRIP = 'DATASTRIP'
    GRANULE = 'GRANULE'
    HTML = 'HTML'
    INFO = 'rep_info'
    QI_DATA = 'QI_DATA'
    IMG_DATA = 'IMG_DATA'
    INSPIRE = 'inspire'
    MANIFEST = 'manifest'
    TCI = 'TCI'
    PVI = 'PVI'
    ECMWFT = 'auxiliary/ECMWFT'
    AUX_ECMWFT = 'auxiliary/AUX_ECMWFT'
    DATASTRIP_METADATA = 'datastrip/*/metadata'

    # More constants about L2A
    AOT = 'AOT'
    WVP = 'WVP'
    SCL = 'SCL'
    VIS = 'VIS'
    L2A_MANIFEST = 'L2AManifest'
    REPORT = 'report'
    GIPP = 'auxiliary/GIP_TL'
    FORMAT_CORRECTNESS = 'FORMAT_CORRECTNESS'
    GENERAL_QUALITY = 'GENERAL_QUALITY'
    GEOMETRIC_QUALITY = 'GEOMETRIC_QUALITY'
    RADIOMETRIC_QUALITY = 'RADIOMETRIC_QUALITY'
    SENSOR_QUALITY = 'SENSOR_QUALITY'
    QUALITY_REPORTS = [FORMAT_CORRECTNESS, GENERAL_QUALITY, GEOMETRIC_QUALITY, RADIOMETRIC_QUALITY, SENSOR_QUALITY]
    CLASS_MASKS = ['SNW', 'CLD']
    R10m = 'R10m'
    R20m = 'R20m'
    R60m = 'R60m'
    RESOLUTIONS = [R10m, R20m, R60m]
    S2_L2A_BAND_MAP = {R10m: ['B02', 'B03', 'B04', 'B08', AOT, TCI, WVP],
                       R20m: ['B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B11', 'B12', AOT, SCL, TCI, VIS, WVP],
                       R60m: ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B09', 'B11', 'B12', AOT, SCL,
                              TCI, WVP]}

    # Order of elements in following lists is important
    # Sentinel-2 L1C products:
    S2_L1C_BANDS = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
    S2_L1C_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, DATASTRIP_METADATA] +\
                       [PREVIEW, PREVIEW_JP2, TCI] +\
                       ['{}/{}'.format(preview, band) for preview, band in it.zip_longest([], S2_L1C_BANDS,
                                                                                          fillvalue=PREVIEW)] +\
                       [QI_MSK_CLOUD] +\
                       ['qi/MSK_{}_{}'.format(qi, band) for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] + \
                       ['qi/{}'.format(qi_report) for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY,
                                                                    GEOMETRIC_QUALITY, SENSOR_QUALITY]] +\
                       [ECMWFT]

    # Sentinel-2 L2A products:
    S2_L2A_BANDS = ['{}/{}'.format(resolution, band) for resolution, band_list in sorted(S2_L2A_BAND_MAP.items())
                    for band in band_list]
    S2_L2A_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, L2A_MANIFEST, REPORT,
                        DATASTRIP_METADATA] + ['datastrip/*/qi/{}'.format(qi_report)for qi_report in QUALITY_REPORTS] +\
                       ['qi/{}_PVI'.format(source_id) for source_id in SOURCE_ID_LIST] +\
                       ['qi/{}_{}'.format(mask, res.lstrip('R')) for mask, res in it.product(CLASS_MASKS,
                                                                                             [R20m, R60m])] +\
                       ['qi/MSK_{}_{}'.format(qi, band) for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] +\
                       [QI_MSK_CLOUD] +\
                       ['qi/{}'.format(qi_report) for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY,
                                                                    GEOMETRIC_QUALITY, SENSOR_QUALITY]] +\
                       [ECMWFT, AUX_ECMWFT, GIPP]

    # Product files with formats:
    PRODUCT_FILES = {**{PRODUCT_INFO: MimeType.JSON,
                        METADATA: MimeType.XML,
                        INSPIRE: MimeType.XML,
                        MANIFEST: MimeType.SAFE,
                        L2A_MANIFEST: MimeType.XML,
                        REPORT: MimeType.XML,
                        DATASTRIP_METADATA: MimeType.XML},
                     **{'datastrip/*/qi/{}'.format(qi_report): MimeType.XML for qi_report in QUALITY_REPORTS}}
    # Tile files with formats:
    TILE_FILES = {**{TILE_INFO: MimeType.JSON,
                     PRODUCT_INFO: MimeType.JSON,
                     METADATA: MimeType.XML,
                     PREVIEW: MimeType.JPG,
                     PREVIEW_JP2: MimeType.JP2,
                     TCI: MimeType.JP2,
                     QI_MSK_CLOUD: MimeType.GML,
                     ECMWFT: MimeType.RAW,
                     AUX_ECMWFT: MimeType.RAW,
                     GIPP: MimeType.XML},
                  **{'qi/{}'.format(qi_report): MimeType.XML for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY,
                                                                               GEOMETRIC_QUALITY, SENSOR_QUALITY]},
                  **{'{}/{}'.format(preview, band): MimeType.JP2
                     for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)},
                  **{'qi/MSK_{}_{}'.format(qi, band): MimeType.GML for qi, band in it.product(QI_LIST, S2_L1C_BANDS)},
                  **{band: MimeType.JP2 for band in S2_L1C_BANDS},
                  **{band: MimeType.JP2 for band in S2_L2A_BANDS},
                  **{'qi/{}_PVI'.format(source_id): MimeType.JP2 for source_id in SOURCE_ID_LIST},
                  **{'qi/{}_{}'.format(mask, res.lstrip('R')): MimeType.JP2 for mask, res in it.product(CLASS_MASKS,
                                                                                                        [R20m, R60m])}}

    # All files joined together
    AWS_FILES = {**PRODUCT_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in PRODUCT_FILES.items()},
                 **TILE_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in TILE_FILES.items()}}


class EsaSafeType(Enum):
    """ Enum constants class for ESA .SAFE type.

     Types are OLD_TYPE and COMPACT_TYPE
    """
    OLD_TYPE = 'old_type'
    COMPACT_TYPE = 'compact_type'



import click

from .data_request import get_safe_format, download_safe_format
from .download import download_data, DownloadRequest
from .config import SHConfig
from .constants import DataSource

# pylint: disable=unused-argument


@click.command()
def main_help():
    """
    Welcome to sentinelhub Python library command line help.

    \b
    There are multiple modules with command line functionality:\n
       - sentinelhub.aws \n
       - senitnelhub.config \n
       - sentinelhub.download \n

    To check more about certain module command use: \n
      sentinelhub.<module name> --help
    """
    pass


@click.command()
@click.option('--product', help='Product ID input')
@click.option('--tile', nargs=2, help='Tile name and date input')
@click.option('-f', '--folder', default='.', help='Set download location')
@click.option('-r', '--redownload', is_flag=True, default=False, help='Redownload existing files')
@click.option('-i', '--info', is_flag=True, default=False, help='Return safe format structure')
@click.option('-e', '--entire', is_flag=True, default=False, help='Get entire product of specified tile')
@click.option('-b', '--bands', default=None, help='Comma separated list (no spaces) of bands to retrieve')
@click.option('--l2a', is_flag=True, default=False, help='In case of tile request this flag specifies L2A products')
def aws(product, tile, folder, redownload, info, entire, bands, l2a):
    """Download Sentinel-2 data from Sentinel-2 on AWS to ESA SAFE format. Download uses multiple threads.

    \b
    Examples with Sentinel-2 L1C data:
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -i
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -f /home/ESA_Products
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 --bands B08,B11
      sentinelhub.aws --tile T54HVH 2017-04-14
      sentinelhub.aws --tile T54HVH 2017-04-14 -e

    \b
    Examples with Sentinel-2 L2A data:
      sentinelhub.aws --product S2A_MSIL2A_20180402T151801_N0207_R068_T33XWJ_20180402T202222
      sentinelhub.aws --tile T33XWJ 2018-04-02 --l2a
    """
    band_list = None if bands is None else bands.split(',')
    data_source = DataSource.SENTINEL2_L2A if l2a else DataSource.SENTINEL2_L1C
    if info:
        if product is None:
            click.echo(get_safe_format(tile=tile, entire_product=entire, data_source=data_source))
        else:
            click.echo(get_safe_format(product_id=product))
    else:
        if product is None:
            download_safe_format(tile=tile, folder=folder, redownload=redownload, entire_product=entire,
                                 bands=band_list, data_source=data_source)
        else:
            download_safe_format(product_id=product, folder=folder, redownload=redownload, bands=band_list)


def config_options(func):
    for param in SHConfig().get_params()[-1::-1]:
        func = click.option('--{}'.format(param), param,
                            help='Set new values to configuration parameter "{}"'.format(param))(func)
    return func


@click.command()
@click.option('--show', is_flag=True, default=False, help='Show current configurations')
@config_options
def config(show, **params):
    """Inspect and configure parameters in your local sentinelhub configuration file

    \b
    Example:
      sentinelhub.config --show
      sentinelhub.config --instance_id <new instance id>
      sentinelhub.config --max_download_attempts 5 --download_sleep_time 20 --download_timeout_seconds 120
    """
    sh_config = SHConfig()
    for param, value in params.items():
        if value is not None:
            try:
                value = int(value)
            except ValueError:
                if value.lower() == 'true':
                    value = True
                elif value.lower() == 'false':
                    value = False
            setattr(sh_config, param, value)
    sh_config.save()

    if show:
        click.echo(str(sh_config))


@click.command()
@click.argument('url')
@click.argument('filename', type=click.Path())
@click.option('-r', '--redownload', is_flag=True, default=False, help='Redownload existing files')
def download(url, filename, redownload):
    """Download from custom created URL into custom created file path

    \b
    Example:
    sentinelhub.download http://sentinel-s2-l1c.s3.amazonaws.com/tiles/54/H/VH/2017/4/14/0/metadata.xml home/example.xml
    """
    data_folder, filename = filename.rsplit('/', 1)
    download_list = [DownloadRequest(url=url, data_folder=data_folder, filename=filename, save_response=True,
                                     return_data=False)]
    download_data(download_list, redownload=redownload)



"""
Module for obtaining data from Amazon Web Service
"""

from abc import ABC, abstractmethod
import logging
import os.path

from .download import DownloadRequest, get_json, AwsDownloadFailedException
from .opensearch import get_tile_info, get_tile_info_id
from .time_utils import parse_time
from .config import SHConfig
from .constants import AwsConstants, EsaSafeType, MimeType, DataSource


LOGGER = logging.getLogger(__name__)


class AwsService(ABC):
    """ Amazon Web Service (AWS) base class

    :param parent_folder: Folder where the fetched data will be saved.
    :type parent_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to ``None`` all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                      If parameter is set to ``None`` the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, parent_folder='', bands=None, metafiles=None):
        self.parent_folder = parent_folder
        self.bands = self._parse_bands(bands)
        self.metafiles = self._parse_metafiles(metafiles)

        self.download_list = []
        self.folder_list = []

    @abstractmethod
    def get_requests(self):
        raise NotImplementedError

    def _parse_bands(self, band_input):
        """
        Parses class input and verifies band names.

        :param band_input: input parameter `bands`
        :type band_input: str or list(str) or None
        :return: verified list of bands
        :rtype: list(str)
        """
        all_bands = AwsConstants.S2_L1C_BANDS if self.data_source is DataSource.SENTINEL2_L1C else \
            AwsConstants.S2_L2A_BANDS

        if band_input is None:
            return all_bands
        if isinstance(band_input, str):
            band_list = band_input.split(',')
        elif isinstance(band_input, list):
            band_list = band_input.copy()
        else:
            raise ValueError('bands parameter must be a list or a string')
        band_list = [band.strip().split('.')[0] for band in band_list]
        band_list = [band for band in band_list if band != '']
        if not set(band_list) <= set(all_bands):
            raise ValueError('bands {} must be a subset of {}'.format(band_list, all_bands))
        return band_list

    def _parse_metafiles(self, metafile_input):
        """
        Parses class input and verifies metadata file names.

        :param metafile_input: class input parameter `metafiles`
        :type metafile_input: str or list(str) or None
        :return: verified list of metadata files
        :rtype: list(str)
        """
        all_metafiles = AwsConstants.S2_L1C_METAFILES if self.data_source is DataSource.SENTINEL2_L1C else \
            AwsConstants.S2_L2A_METAFILES

        if metafile_input is None:
            if self.__class__.__name__ == 'SafeProduct':
                return all_metafiles
            elif self.__class__.__name__ == 'SafeTile':
                return [metafile for metafile in all_metafiles if metafile in AwsConstants.TILE_FILES]
            return []
        if isinstance(metafile_input, str):
            metafile_list = metafile_input.split(',')
        elif isinstance(metafile_input, list):
            metafile_list = metafile_input.copy()
        else:
            raise ValueError('metafiles parameter must be a list or a string')
        metafile_list = [metafile.strip().split('.')[0] for metafile in metafile_list]
        metafile_list = [metafile for metafile in metafile_list if metafile != '']
        if not set(metafile_list) <= set(all_metafiles):
            raise ValueError('metadata files {} must be a subset of {}'.format(metafile_list, all_metafiles))
        return metafile_list

    def get_base_url(self):
        if self.data_source is DataSource.SENTINEL2_L1C:
            if SHConfig().use_s3_l1c_bucket:
                return 's3://{}/'.format(SHConfig().aws_s3_l1c_bucket)
            return SHConfig().aws_base_url
        return 's3://{}/'.format(SHConfig().aws_s3_l2a_bucket)

    def get_safe_type(self):
        """Determines the type of ESA product.

        In 2016 ESA changed structure and naming of data. Therefore the class must
        distinguish between old product type and compact (new) product type.

        :return: type of ESA product
        :rtype: constants.EsaSafeType
        :raises: ValueError
        """
        product_type = self.product_id.split('_')[1]
        if product_type.startswith('MSI'):
            return EsaSafeType.COMPACT_TYPE
        if product_type in ['OPER', 'USER']:
            return EsaSafeType.OLD_TYPE
        raise ValueError('Unrecognized product type of product id {}'.format(self.product_id))

    def get_baseline(self):
        """ Determines the baseline number (i.e. version) of ESA .SAFE product

        :return: baseline number
        :rtype: str
        :raises: ValueError
        """
        if self.safe_type is EsaSafeType.COMPACT_TYPE:
            baseline = self.product_id.split('_')[3].lstrip('N')
            if len(baseline) != 4:
                raise ValueError('Unable to recognize baseline number from the product id {}'.format(self.product_id))
            return '{}.{}'.format(baseline[:2], baseline[2:])
        return self._read_baseline_from_info()

    def _read_baseline_from_info(self):
        """Tries to find and return baseline number from either tileInfo or productInfo file.

        :return: Baseline ID
        :rtype: str
        :raises: ValueError
        """
        if hasattr(self, 'tile_info'):
            return self.tile_info['datastrip']['id'][-5:]
        if hasattr(self, 'product_info'):
            return self.product_info['datastrips'][0]['id'][-5:]
        raise ValueError('No info file has been obtained yet.')

    @staticmethod
    def url_to_tile(url):
        """
        Extracts tile name, date and AWS index from tile url on AWS.

        :param url: class input parameter 'metafiles'
        :type url: str
        :return: Name of tile, date and AWS index which uniquely identifies tile on AWS
        :rtype: (str, str, int)
        """
        info = url.strip('/').split('/')
        name = ''.join(info[-7: -4])
        date = '-'.join(info[-4: -1])
        return name, date, int(info[-1])

    def sort_download_list(self):
        """
        Method for sorting the list of download requests. Band images have priority before metadata files. If bands
        images or metadata files are specified with a list they will be sorted in the same order as in the list.
        Otherwise they will be sorted alphabetically (band B8A will be between B08 and B09).
        """
        def aws_sort_function(download_request):
            data_name = download_request.properties['data_name']
            url_part = download_request.url[len(self.base_url):].strip('/')
            url_part = ''.join(url_part.split('/', 4)[:-1])
            if data_name in self.bands:
                return 0, url_part, self.bands.index(data_name)
            return 1, url_part, self.metafiles.index(data_name)
        self.download_list.sort(key=aws_sort_function)

    def structure_recursion(self, struct, folder):
        """
        From nested dictionaries representing .SAFE structure it recursively extracts all the files that need to be
        downloaded and stores them into class attribute `download_list`.

        :param struct: nested dictionaries representing a part of .SAFE structure
        :type struct: dict
        :param folder: name of folder where this structure will be saved
        :type folder: str
        """
        has_subfolder = False
        for name, substruct in struct.items():
            subfolder = os.path.join(folder, name)
            if not isinstance(substruct, dict):
                if substruct.split('/')[3] == 'products':
                    data_name = substruct.split('/', 8)[-1]
                    if data_name.startswith('datastrip'):
                        items = data_name.split('/', 2)
                        data_name = '/'.join([items[0], '*', items[2]])
                else:
                    data_name = substruct.split('/', 11)[-1]
                if '.' in data_name:
                    data_type = MimeType(substruct.split('.')[-1])
                    data_name = data_name.rsplit('.', 1)[0]
                else:
                    data_type = MimeType.RAW
                if data_name in self.bands + self.metafiles:
                    self.download_list.append(DownloadRequest(url=substruct, filename=subfolder, data_type=data_type,
                                                              data_name=data_name))
            else:
                has_subfolder = True
                self.structure_recursion(substruct, subfolder)
        if not has_subfolder:
            self.folder_list.append(folder)

    @staticmethod
    def add_file_extension(filename, data_format=None, remove_path=False):
        """Joins filename and corresponding file extension if it has one.

        :param filename: Name of the file without extension
        :type filename: str
        :param data_format: format of file, if None it will be set automatically
        :type data_format: constants.MimeType or None
        :param remove_path: True if the path in filename string should be removed
        :type remove_path: bool
        :return: Name of the file with extension
        :rtype: str
        """
        if data_format is None:
            data_format = AwsConstants.AWS_FILES[filename]
        if remove_path:
            filename = filename.split('/')[-1]
        if filename.startswith('datastrip'):
            filename = filename.replace('*', '0')
        if data_format is MimeType.RAW:
            return filename
        return '{}.{}'.format(filename.replace('*', ''), data_format.value)

    def has_reports(self):
        """Products created with baseline 2.06 and greater (and some products with baseline 2.05) should have quality
        report files

        :return: True if the product has report xml files and False otherwise
        :rtype: bool
        """
        return self.baseline > '02.05' or (self.baseline == '02.05' and self.date >= '2017-10-12')  # Not sure

    def is_early_compact_l2a(self):
        """Check if product is early version of compact L2A product

        :return: True if product is early version of compact L2A product and False otherwise
        :rtype: bool
        """
        return self.data_source is DataSource.SENTINEL2_L2A and self.safe_type is EsaSafeType.COMPACT_TYPE and \
            self.baseline <= '02.06'


class AwsProduct(AwsService):
    """ Service class for Sentinel-2 product on AWS

    :param product_id: ESA ID of the product
    :type product_id: str
    :param tile_list: list of tile names
    :type tile_list: list(str) or None
    :param parent_folder: location of the directory where the fetched data will be saved.
    :type parent_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to ``None`` all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                      If parameter is set to ``None`` the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, product_id, tile_list=None, **kwargs):
        self.product_id = product_id.split('.')[0]
        self.tile_list = self.parse_tile_list(tile_list)

        self.data_source = self.get_data_source()
        self.safe_type = self.get_safe_type()
        self.base_url = self.get_base_url()

        self.date = self.get_date()
        self.product_url = self.get_product_url()
        self.product_info = get_json(self.get_url(AwsConstants.PRODUCT_INFO))

        self.baseline = self.get_baseline()

        super(AwsProduct, self).__init__(**kwargs)

    @staticmethod
    def parse_tile_list(tile_input):
        """
        Parses class input and verifies band names.

        :param tile_input: class input parameter `tile_list`
        :type tile_input: str or list(str)
        :return: parsed list of tiles
        :rtype: list(str) or None
        """
        if tile_input is None:
            return None
        if isinstance(tile_input, str):
            tile_list = tile_input.split(',')
        elif isinstance(tile_input, list):
            tile_list = tile_input.copy()
        else:
            raise ValueError('tile_list parameter must be a list of tile names')
        tile_list = [AwsTile.parse_tile_name(tile_name) for tile_name in tile_list]
        return tile_list

    def get_requests(self):
        """
        Creates product structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        :rtype: (list(download.DownloadRequest), list(str))
        """
        self.download_list = [DownloadRequest(url=self.get_url(metafile), filename=self.get_filepath(metafile),
                                              data_type=AwsConstants.AWS_FILES[metafile], data_name=metafile) for
                              metafile in self.metafiles if metafile in AwsConstants.PRODUCT_FILES]

        tile_parent_folder = os.path.join(self.parent_folder, self.product_id)
        for tile_info in self.product_info['tiles']:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_downloads, tile_folders = AwsTile(tile_name, date, aws_index, parent_folder=tile_parent_folder,
                                                       bands=self.bands, metafiles=self.metafiles,
                                                       data_source=self.data_source).get_requests()
                self.download_list.extend(tile_downloads)
                self.folder_list.extend(tile_folders)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_data_source(self):
        """The method determines data source from product ID.

        :return: Data source of the product
        :rtype: DataSource
        :raises: ValueError
        """
        product_type = self.product_id.split('_')[1]
        if product_type.endswith('L1C') or product_type == 'OPER':
            return DataSource.SENTINEL2_L1C
        if product_type.endswith('L2A') or product_type == 'USER':
            return DataSource.SENTINEL2_L2A
        raise ValueError('Unknown data source of product {}'.format(self.product_id))

    def get_date(self):
        """ Collects sensing date of the product.

        :return: Sensing date
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = self.product_id.split('_')[-2]
            date = [name[1:5], name[5:7], name[7:9]]
        else:
            name = self.product_id.split('_')[2]
            date = [name[:4], name[4:6], name[6:8]]
        return '-'.join(date_part.lstrip('0') for date_part in date)

    def get_url(self, filename, data_format=None):
        """
        Creates url of file location on AWS.

        :param filename: name of file
        :type filename: str
        :param data_format: format of file, if None it will be set automatically
        :type data_format: constants.MimeType or None
        :return: url of file location
        :rtype: str
        """
        if self.product_url is None:
            self.product_url = self.get_product_url()
        return '{}/{}'.format(self.product_url, self.add_file_extension(filename, data_format))

    def get_product_url(self):
        """
        Creates base url of product location on AWS.

        :return: url of product location
        :rtype: str
        """
        return '{}products/{}/{}'.format(self.base_url, self.date.replace('-', '/'), self.product_id)

    def get_tile_url(self, tile_info):
        """
        Collects tile url from productInfo.json file.

        :param tile_info: information about tile from productInfo.json
        :type tile_info: dict
        :return: url of tile location
        :rtype: str
        """
        return '{}/{}'.format(self.base_url, tile_info['path'])

    def get_filepath(self, filename):
        """
        Creates file path for the file.

        :param filename: name of the file
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return os.path.join(self.parent_folder, self.product_id, self.add_file_extension(filename)).replace(':', '.')


class AwsTile(AwsService):
    """ Service class for Sentinel-2 product on AWS

    :param tile: Tile name (e.g. 'T10UEV')
    :type tile: str
    :param time: Tile sensing time in ISO8601 format
    :type time: str
    :param aws_index: There exist Sentinel-2 tiles with the same tile and time parameter. Therefore each tile on AWS
                      also has an index which is visible in their url path. If ``aws_index`` is set to ``None`` the
                      class will try to find the index automatically. If there will be multiple choices it will choose
                      the lowest index and inform the user.
    :type aws_index: int or None
    :param data_source: Source of requested AWS data. Supported sources are Sentinel-2 L1C and Sentinel-2 L2A, default
                        is Sentinel-2 L1C data.
    :type data_source: constants.DataSource
    :param parent_folder: folder where the fetched data will be saved.
    :type parent_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to ``None`` all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                      If parameter is set to ``None`` the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, tile_name, time, aws_index=None, data_source=DataSource.SENTINEL2_L1C, **kwargs):
        self.tile_name = self.parse_tile_name(tile_name)
        self.datetime = self.parse_datetime(time)
        self.date = self.datetime.split('T')[0]
        self.aws_index = aws_index
        self.data_source = data_source

        self.base_url = self.get_base_url()

        self.aws_index = self.get_aws_index()
        self.tile_url = self.get_tile_url()
        self.tile_info = self.get_tile_info()
        if not self.tile_is_valid():
            raise ValueError('Cannot find data on AWS for specified tile, time and aws_index')

        self.product_id = self.get_product_id()
        self.safe_type = self.get_safe_type()
        self.baseline = self.get_baseline()

        super(AwsTile, self).__init__(**kwargs)

    @staticmethod
    def parse_tile_name(name):
        """
        Parses and verifies tile name.

        :param name: class input parameter `tile_name`
        :type name: str
        :return: parsed tile name
        :rtype: str
        """
        tile_name = name.lstrip('T0')
        if len(tile_name) == 4:
            tile_name = '0' + tile_name
        if len(tile_name) != 5:
            raise ValueError('Invalid tile name {}'.format(name))
        return tile_name

    @staticmethod
    def parse_datetime(time):
        """
        Parses and verifies tile sensing time.

        :param time: tile sensing time
        :type time: str
        :return: tile sensing time in ISO8601 format
        :rtype: str
        """
        try:
            return parse_time(time)
        except Exception:
            raise ValueError('Time must be in format YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS')

    def get_requests(self):
        """
        Creates tile structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        :rtype: (list(download.DownloadRequest), list(str))
        """
        self.download_list = []
        for data_name in [band for band in self.bands if self._band_exists(band)] + self.metafiles:
            if data_name in AwsConstants.TILE_FILES:
                url = self.get_url(data_name)
                filename = self.get_filepath(data_name)
                self.download_list.append(DownloadRequest(url=url, filename=filename,
                                                          data_type=AwsConstants.AWS_FILES[data_name],
                                                          data_name=data_name))
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_aws_index(self):
        """
        Returns tile index on AWS. If `tile_index` was not set during class initialization it will be determined
        according to existing tiles on AWS.

        :return: Index of tile on AWS
        :rtype: int
        """
        if self.aws_index is not None:
            return self.aws_index
        tile_info_list = get_tile_info(self.tile_name, self.datetime, all_tiles=True)
        if not tile_info_list:
            raise ValueError('Cannot find aws_index for specified tile and time')

        if self.data_source is DataSource.SENTINEL2_L2A:
            for tile_info in sorted(tile_info_list, key=self._parse_aws_index):
                try:
                    self.aws_index = self._parse_aws_index(tile_info)
                    self.get_tile_info()
                    return self.aws_index
                except AwsDownloadFailedException:
                    pass

        return self._parse_aws_index(tile_info_list[0])

    @staticmethod
    def _parse_aws_index(tile_info):
        """Parses AWS index from tile info

        :param tile_info: dictionary with information about tile
        :type tile_info: dict
        :return: Index of tile on AWS
        :rtype: int
        """
        return int(tile_info['properties']['s3Path'].split('/')[-1])

    def tile_is_valid(self):
        return self.tile_info is not None \
               and (self.datetime == self.date or self.datetime == self.parse_datetime(self.tile_info['timestamp']))

    def get_tile_info(self):
        """
        Collects basic info about tile from tileInfo.json.

        :return: dictionary with tile information
        :rtype: dict
        """
        url = self.get_url(AwsConstants.TILE_INFO)
        try:
            return get_json(url)
        except Exception as err:
            LOGGER.error('Download from url %s failed with %s', url, err)
            raise

    def get_url(self, filename):
        """
        Creates url of file location on AWS.

        :param filename: name of file
        :type filename: str
        :return: url of file location
        :rtype: str
        """
        if not hasattr(self, 'tile_url') or filename == AwsConstants.TILE_INFO:
            self.tile_url = self.get_tile_url()
        return '{}/{}'.format(self.tile_url, self.add_file_extension(filename))

    def get_tile_url(self):
        """
        Creates base url of tile location on AWS.

        :return: url of tile location
        :rtype: str
        """
        url = '{}tiles/{}/{}/{}/'.format(self.base_url, self.tile_name[0:2].lstrip('0'), self.tile_name[2],
                                         self.tile_name[3:5])
        date_params = self.date.split('-')
        for param in date_params:
            url += param.lstrip('0') + '/'
        return url + str(self.aws_index)

    def get_qi_url(self, metafile):
        """Returns url of tile metadata products

        :param metafile: Name of metadata product at AWS
        :type metafile: str
        :return: url location of metadata product at AWS
        :rtype: str
        """
        return '{}/qi/{}'.format(self.tile_url, metafile)

    def get_gml_url(self, qi_type, band='B00'):
        """
        :param qi_type: type of quality indicator
        :type qi_type: str
        :param band: band name
        :type band: str
        :return: location of gml file on AWS
        :rtype: str
        """
        band = band.split('/')[-1]
        return self.get_qi_url('MSK_{}_{}.gml'.format(qi_type, band))

    def get_preview_url(self, data_type='L1C'):
        """Returns url location of full resolution L1C preview
        :return:
        """
        if self.data_source is DataSource.SENTINEL2_L1C or self.safe_type is EsaSafeType.OLD_TYPE:
            return self.get_url(AwsConstants.PREVIEW_JP2)
        return self.get_qi_url('{}_PVI.jp2'.format(data_type))

    def get_filepath(self, filename):
        """
        Creates file path for the file.

        :param filename: name of the file
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return os.path.join(self.parent_folder, '{},{},{}'.format(self.tile_name, self.date, self.aws_index),
                            self.add_file_extension(filename)).replace(':', '.')

    def get_product_id(self):
        """
        Obtains ESA ID of product which contains the tile.

        :return: ESA ID of the product
        :rtype: str
        """
        return self.tile_info['productName']

    def _band_exists(self, band_name):
        if self.data_source is DataSource.SENTINEL2_L1C:
            return True
        resolution, band = band_name.split('/')
        if self.safe_type is EsaSafeType.COMPACT_TYPE:
            return not (self.baseline >= '02.07' and band.endswith(AwsConstants.VIS))
        return band != AwsConstants.TCI and not (band == AwsConstants.SCL and resolution == AwsConstants.R60m)

    @staticmethod
    def tile_id_to_tile(tile_id):
        """
        :param tile_id: original tile identification string provided by ESA (e.g.
                        'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
        :type: str
        :return: tile name, sensing date and AWS index
        :rtype: (str, str, int)
        """
        if tile_id.split('_')[0] not in ['S2A', 'L1C']:
            raise ValueError('Transformation from tile_id to tile works currently only for Sentinel-2 L1C products')

        tile_info = get_tile_info_id(tile_id)
        return AwsService.url_to_tile(tile_info['properties']['s3Path'])



import io
import os
from setuptools import setup, find_packages


def parse_requirements(file):
    return sorted(set(
        line.partition('#')[0].strip()
        for line in open(os.path.join(os.path.dirname(__file__), file))
    ) - set(''))


def get_version():
    for line in open(os.path.join(os.path.dirname(__file__), 'sentinelhub', '__init__.py')):
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"').strip("'")
    return version


def get_long_description():
    return io.open('README.md', encoding="utf-8").read()


setup(
    name='sentinelhub',
    python_requires='>=3.5,<3.7',
    version=get_version(),
    description='Sentinel Hub Utilities',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url='https://github.com/sentinel-hub/sentinelhub-py',
    author='Sinergise ltd.',
    author_email='info@sentinel-hub.com',
    license='MIT',
    packages=find_packages(),
    package_data={'sentinelhub': ['sentinelhub/config.json']},
    include_package_data=True,
    install_requires=parse_requirements("requirements.txt"),
    extras_require={'DEV': parse_requirements("requirements-dev.txt")},
    zip_safe=False,
    entry_points={'console_scripts': ['sentinelhub=sentinelhub.commands:main_help',
                                      'sentinelhub.aws=sentinelhub.commands:aws',
                                      'sentinelhub.config=sentinelhub.commands:config',
                                      'sentinelhub.download=sentinelhub.commands:download']}
)



"""
Module for working with Geopedia OGC services
"""

import logging

from .ogc import OgcImageService
from .config import SHConfig

LOGGER = logging.getLogger(__name__)


class GeopediaService:
    """ The class for Geopedia OGC services

    :param base_url: Base url of Geopedia's OGC services. If ``None``, the url specified in the configuration
                    file is taken.
    :type base_url: str or None
    """
    def __init__(self, base_url=None):
        self.base_url = SHConfig().gpd_base_url if base_url is None else base_url


class GeopediaImageService(GeopediaService, OgcImageService):
    """Geopedia OGC services class for providing image data. Most of the methods are inherited from
    `sentinelhub.ogc.OgcImageService` class.

    :param base_url: Base url of Geopedia's OGC services. If ``None``, the url specified in the configuration
                    file is taken.
    :type base_url: str or None
    """
    def __init__(self, **kwargs):
        super(GeopediaImageService, self).__init__(**kwargs)

    def get_dates(self, request):
        """ Geopedia does not support date queries

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: Undefined date
        :rtype: [None]
        """
        return [None]

    def get_wfs_iterator(self):
        """ This method is inherited from OgcImageService but is not implemented.
        """
        raise NotImplementedError



"""
Module for manipulation of geographical information.
"""

# pylint: disable=unused-argument

import logging
import pyproj

from .constants import CRS
from .common import BBox


LOGGER = logging.getLogger(__name__)

ERR = 0.1


def bbox_to_resolution(bbox, width, height):
    """ Calculates pixel resolution in meters for a given bbox of a given width and height.

    :param bbox: bounding box
    :type bbox: common.BBox
    :param width: width of bounding box in pixels
    :type width: int
    :param height: height of bounding box in pixels
    :type height: int
    :return: resolution east-west at north and south, and resolution north-south in meters for given CRS
    :rtype: float, float, float
    :raises: ValueError if CRS is not supported
    """
    utm_bbox = to_utm_bbox(bbox)
    east1, north1 = utm_bbox.get_lower_left()
    east2, north2 = utm_bbox.get_upper_right()
    return abs(east2 - east1) / width, abs(north2 - north1) / height


def get_image_dimension(bbox, width=None, height=None):
    """ Given bounding box and one of the parameters width or height it will return the other parameter that will best
    fit the bounding box dimensions

    :param bbox: bounding box
    :type bbox: common.BBox
    :param width: image width or None if height is unknown
    :type width: int or None
    :param height: image height or None if height is unknown
    :type height: int or None
    :return: width or height rounded to integer
    :rtype: int
    """
    utm_bbox = to_utm_bbox(bbox)
    east1, north1 = utm_bbox.get_lower_left()
    east2, north2 = utm_bbox.get_upper_right()
    if isinstance(width, int):
        return round(width * abs(north2 - north1) / abs(east2 - east1))
    return round(height * abs(east2 - east1) / abs(north2 - north1))


def to_utm_bbox(bbox):
    """ Transform bbox into UTM CRS

    :param bbox: bounding box
    :type bbox: common.BBox
    :return: bounding box in UTM CRS
    :rtype: common.BBox
    """
    if CRS.is_utm(bbox.get_crs()):
        return bbox
    lng, lat = bbox.get_middle()
    utm_crs = get_utm_crs(lng, lat, source_crs=bbox.get_crs())
    return transform_bbox(bbox, utm_crs)


def get_utm_bbox(img_bbox, transform):
    """ Get UTM coordinates given a bounding box in pixels and a transform

    :param img_bbox: boundaries of bounding box in pixels as [row1, col1, row2, col2]
    :type img_bbox: list
    :param transform: georeferencing transform of the image, e.g. (x_upper_left, res_x, 0, y_upper_left, 0, -res_y)
    :type transform: tuple or list
    :return: UTM coordinates as [east1, north1, east2, north2]
    :rtype: list
    """
    east1, north1 = pixel_to_utm(img_bbox[0], img_bbox[1], transform)
    east2, north2 = pixel_to_utm(img_bbox[2], img_bbox[3], transform)
    return [east1, north1, east2, north2]


def wgs84_to_utm(lng, lat, utm_crs=None):
    """ Convert WGS84 coordinates to UTM. If UTM CRS is not set it will be calculated automatically.

    :param lng: longitude in WGS84 system
    :type lng: float
    :param lat: latitude in WGS84 system
    :type lat: float
    :param utm_crs: UTM coordinate reference system enum constants
    :type utm_crs: constants.CRS or None
    :return: east, north coordinates in UTM system
    :rtype: float, float
    """
    if utm_crs is None:
        utm_crs = get_utm_crs(lng, lat)
    return transform_point((lng, lat), CRS.WGS84, utm_crs)


def to_wgs84(east, north, crs):
    """ Convert any CRS with (east, north) coordinates to WGS84

    :param east: east coordinate
    :type east: float
    :param north: north coordinate
    :type north: float
    :param crs: CRS enum constants
    :type crs: constants.CRS
    :return: latitude and longitude coordinates in WGS84 system
    :rtype: float, float
    """
    return transform_point((east, north), crs, CRS.WGS84)


def utm_to_pixel(east, north, transform, truncate=True):
    """ Convert UTM coordinate to image coordinate given a transform

    :param east: east coordinate of point
    :type east: float
    :param north: north coordinate of point
    :type north: float
    :param transform: georeferencing transform of the image, e.g. (x_upper_left, res_x, 0, y_upper_left, 0, -res_y)
    :type transform: tuple or list
    :param truncate: Whether to truncate pixel coordinates. Default is ``True``
    :type truncate: bool
    :return: row and column pixel image coordinates
    :rtype: float, float or int, int
    """
    column = (east - transform[0]) / transform[1]
    row = (north - transform[3]) / transform[5]
    if truncate:
        return int(row + ERR), int(column + ERR)
    return row, column


def pixel_to_utm(row, column, transform):
    """ Convert pixel coordinate to UTM coordinate given a transform

    :param row: row pixel coordinate
    :type row: int or float
    :param column: column pixel coordinate
    :type column: int or float
    :param transform: georeferencing transform of the image, e.g. (x_upper_left, res_x, 0, y_upper_left, 0, -res_y)
    :type transform: tuple or list
    :return: east, north UTM coordinates
    :rtype: float, float
    """
    east = transform[0] + column * transform[1]
    north = transform[3] + row * transform[5]
    return east, north


def wgs84_to_pixel(lng, lat, transform, utm_epsg=None, truncate=True):
    """ Convert WGS84 coordinates to pixel image coordinates given transform and UTM CRS. If no CRS is given it will be
    calculated it automatically.

    :param lng: longitude of point
    :type lng: float
    :param lat: latitude of point
    :type lat: float
    :param transform: georeferencing transform of the image, e.g. (x_upper_left, res_x, 0, y_upper_left, 0, -res_y)
    :type transform: tuple or list
    :param utm_epsg: UTM coordinate reference system enum constants
    :type utm_epsg: constants.CRS or None
    :param truncate: Whether to truncate pixel coordinates. Default is ``True``
    :type truncate: bool
    :return: row and column pixel image coordinates
    :rtype: float, float or int, int
    """
    east, north = wgs84_to_utm(lng, lat, utm_epsg)
    row, column = utm_to_pixel(east, north, transform, truncate=truncate)
    return row, column


def get_utm_crs(lng, lat, source_crs=CRS.WGS84):
    """ Get CRS for UTM zone in which (lat, lng) is contained.

    :param lng: longitude
    :type lng: float
    :param lat: latitude
    :type lat: float
    :param source_crs: source CRS
    :type source_crs: constants.CRS
    :return: CRS of the zone containing the lat,lon point
    :rtype: constants.CRS
    """
    if source_crs is not CRS.WGS84:
        lng, lat = transform_point((lng, lat), source_crs, CRS.WGS84)
    return CRS.get_utm_from_wgs84(lng, lat)


def transform_point(point, source_crs, target_crs):
    """ Maps point form src_crs to tgt_crs

    :param point: a tuple (x, y)
    :type point: (float, float)
    :param source_crs: source CRS
    :type source_crs: constants.CRS
    :param target_crs: target CRS
    :type target_crs: constants.CRS
    :return: point in target CRS
    :rtype: (float, float)
    """
    if source_crs == target_crs:
        return point
    old_x, old_y = point
    new_x, new_y = pyproj.transform(CRS.projection(source_crs), CRS.projection(target_crs), old_x, old_y)
    return new_x, new_y


def transform_bbox(bbox, target_crs):
    """ Maps bbox from current crs to target_crs

    :param bbox: bounding box
    :type bbox: common.BBox
    :param target_crs: target CRS
    :type target_crs: constants.CRS
    :return: bounding box in target CRS
    :rtype: common.BBox
    """
    source_crs = bbox.get_crs()
    lower_left = bbox.get_lower_left()
    upper_right = bbox.get_upper_right()
    new_lower_left = transform_point(lower_left, source_crs, target_crs)
    new_upper_right = transform_point(upper_right, source_crs, target_crs)
    return BBox((new_lower_left, new_upper_right), target_crs)
